# =============================================================================
# Minet FileSystem Utilities
# =============================================================================
#
# Multiple helper functions related to reading and writing files.
#
from typing import Union, Optional

import gzip
import json
import yaml
from ebbe.decorators import with_defer
from os import makedirs, PathLike
from os.path import basename, join, splitext, abspath, normpath, dirname
from ural import get_hostname, get_normalized_hostname
from quenouille import NamedLocks

from minet.exceptions import (
    FilenameFormattingError,
    DefinitionInvalidFormatError,
    CouldNotInferEncodingError,
)
from minet.utils import md5, PseudoFStringFormatter
from minet.encodings import infer_encoding


def read_potentially_gzipped_path(
    path, encoding: Optional[str] = None, errors: str = "replace"
) -> str:
    open_fn = open
    flag = "r" if encoding is not None else "rb"

    if path.endswith(".gz"):
        open_fn = gzip.open
        flag = "rt" if encoding is not None else "r"

    if encoding is None:
        with open_fn(path, flag) as f:
            binary = f.read()
            encoding = infer_encoding(binary)

            if encoding is None:
                raise CouldNotInferEncodingError

            return binary.decode(encoding, errors=errors)

    with open_fn(path, flag, encoding=encoding, errors=errors) as f:
        return f.read()


@with_defer()
def load_definition(f, *, defer=None, encoding="utf-8"):
    if isinstance(f, (str, PathLike)):
        path = str(f)
        f = open(path, encoding=encoding)
        defer(f.close)  # type: ignore
    else:
        path = f.name

    if path.endswith(".json"):
        definition = json.load(f)

    elif path.endswith(".yml") or path.endswith(".yaml"):
        definition = yaml.safe_load(f)

    else:
        raise DefinitionInvalidFormatError

    return definition


class FolderStrategy(object):
    def __call__(self):
        raise NotImplementedError

    @staticmethod
    def from_name(name):
        if name == "flat":
            return FlatFolderStrategy()

        if name == "hostname":
            return HostnameFolderStrategy()

        if name == "normalized-hostname":
            return NormalizedHostnameFolderStrategy()

        if name.startswith("prefix-"):
            length = name.split("prefix-")[-1]

            try:
                length = int(length)
            except ValueError:
                raise TypeError

            if length <= 0:
                raise TypeError

            return PrefixFolderStrategy(length)

        raise TypeError


class FlatFolderStrategy(FolderStrategy):
    def __call__(self, filename, **kwargs):
        return filename


class PrefixFolderStrategy(FolderStrategy):
    def __init__(self, length):
        self.length = length

    def __call__(self, filename, **kwargs):
        base = basename(filename).split(".", 1)[0]

        return join(base[: self.length], filename)


class HostnameFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        hostname = get_hostname(url)

        if not hostname:
            hostname = "unknown-host"

        return join(hostname, filename)


class NormalizedHostnameFolderStrategy(FolderStrategy):
    def __call__(self, filename, url, **kwargs):
        hostname = get_normalized_hostname(
            url,
            normalize_amp=False,
            strip_lang_subdomains=True,
            infer_redirection=False,
        )

        if not hostname:
            hostname = "unknown-host"

        return join(hostname, filename)


class FilenameBuilder(object):
    def __init__(self, folder_strategy=None, template=None):
        self.folder_strategy = None

        if folder_strategy is not None:
            self.folder_strategy = FolderStrategy.from_name(folder_strategy)

        self.formatter = PseudoFStringFormatter()
        self.template = template

    def __call__(
        self, url=None, filename=None, ext=None, formatter_kwargs={}, compressed=False
    ):
        original_ext = None

        if filename is None:
            base = md5(url)
        else:
            base, original_ext = splitext(filename)

        # We favor the extension found in given filename, else we fallback
        # on the provided one if any (usually inferred from http response)
        ext = original_ext if original_ext else (ext or "")

        if self.template is not None:
            try:
                filename = self.formatter.format(
                    self.template, value=base, ext=ext, **formatter_kwargs
                )
            except Exception as e:
                raise FilenameFormattingError(reason=e, template=self.template)
        else:
            filename = base + ext

        if self.folder_strategy:
            filename = self.folder_strategy(filename, url=url)

        if compressed:
            filename += ".gz"

        return filename


class ThreadSafeFilesWriter(object):
    def __init__(self, root_directory=""):
        self.root_directory = root_directory
        self.folder_locks = NamedLocks()
        self.file_locks = NamedLocks()

    def resolve(self, filename, relative=False):
        full_path = join(self.root_directory, filename)

        if relative:
            return normpath(full_path)

        return abspath(full_path)

    def makedirs(self, directory):
        if not directory:
            return

        # TODO: cache
        with self.folder_locks[directory]:
            makedirs(directory, exist_ok=True)

    def write(self, filename: str, contents: Union[str, bytes], compress: bool = False):
        binary = isinstance(contents, bytes)

        if compress and not binary:
            raise NotImplementedError

        filename = self.resolve(filename)
        directory = dirname(filename)

        # NOTE: Could have prefix-free locking as a bonus...
        self.makedirs(directory)

        open_kwargs = {"mode": "wb" if binary else "w"}

        if not binary:
            open_kwargs["encoding"] = "utf-8"

        if compress:
            open_fn = gzip.open
        else:
            open_fn = open

        with self.file_locks[filename]:
            with open_fn(filename, **open_kwargs) as f:
                f.write(contents)



from typing import Generic, TypeVar, Optional, Any, Iterable, List, Tuple
from minet.types import Literal

from shutil import rmtree
from os.path import isfile, join
from queue import Queue, LifoQueue as Stack, Empty
from persistqueue import SQLiteAckQueue, FILOSQLiteQueue as SQLiteAckStack
from threading import Lock

ItemType = TypeVar("ItemType")

DB_FILE_NAME = "queue.db"

ACK_STATUS_TO_NAME = {
    0: "inited",
    1: "ready",
    2: "unack",
    5: "acked",
    9: "acked_failed",
}

AckStatusType = Literal["inited", "ready", "unack", "acked", "acked_failed"]
DumpedItemType = Tuple[AckStatusType, ItemType]
DumpType = List[DumpedItemType[ItemType]]


class CrawlerQueue(Generic[ItemType]):
    path: Optional[str]
    persistent: bool
    resuming: bool
    cleanup_interval: int

    __write_lock: Lock
    __queue: Any
    __current_task_done_count: int

    def __init__(
        self,
        path: Optional[str] = None,
        resume: bool = False,
        cleanup_interval: int = 5000,
        dfs: bool = False,
    ):
        self.path = path
        self.resuming = False
        self.persistent = False
        self.cleanup_interval = cleanup_interval

        self.__write_lock = Lock()

        if path is not None:
            self.persistent = True

            if not resume:
                rmtree(path, ignore_errors=True)
            else:
                if isfile(join(path, DB_FILE_NAME)):
                    self.resuming = True

            QueueCls = SQLiteAckStack if dfs else SQLiteAckQueue

            self.__queue = QueueCls(
                path, db_file_name=DB_FILE_NAME, multithreading=True, auto_resume=True
            )

        else:
            self.__queue = Stack() if dfs else Queue()

        self.__current_task_done_count = 0

    def qsize(self) -> int:
        return self.__queue.qsize()

    def put(self, item: ItemType) -> None:
        with self.__write_lock:
            self.__queue.put(item)

    def put_many(self, items: Iterable[ItemType]) -> int:
        with self.__write_lock:
            count = 0

            for item in items:
                self.__queue.put(item)
                count += 1

            return count

    def get(self, block=True) -> ItemType:
        return self.__queue.get(block=block)

    def get_nowait(self) -> ItemType:
        return self.get(False)

    def ack(self, item: ItemType) -> None:
        self.__current_task_done_count += 1

        if self.persistent:
            self.__queue.ack(item)

        self.__queue.task_done()

        if self.__current_task_done_count >= self.cleanup_interval:
            self.cleanup()

    def cleanup(self):
        with self.__write_lock:
            if not self.persistent:
                return

            self.__current_task_done_count = 0

            self.__queue.clear_acked_data(
                max_delete=None, keep_latest=None, clear_ack_failed=False
            )
            self.__queue.shrink_disk_usage()

    def dump(self) -> DumpType[ItemType]:
        with self.__write_lock:
            if not self.persistent:
                items = []

                while True:
                    try:
                        items.append(("unack", self.__queue.get_nowait()))
                    except Empty:
                        break

                for item in items:
                    self.__queue.put_nowait(item)

                return items

            records = self.__queue.queue()

            items = []

            for record in records:
                items.append((ACK_STATUS_TO_NAME[record["status"]], record["data"]))

            return items

    def __del__(self):
        del self.__queue



# =============================================================================
# Minet Twitter Tweet-date CLI Action
# =============================================================================
#
# Logic of the `tw tweet-date` action.
#
import re

from minet.cli.utils import with_enricher_and_loading_bar
from twitwi.utils import get_dates_from_id
from ural.twitter import parse_twitter_url, TwitterTweet, TwitterUser, TwitterList

ID_RE = re.compile(r"^[0-9]+$")


@with_enricher_and_loading_bar(
    headers=["timestamp_utc", "locale_time"],
    title="Inferring tweet dates",
    unit="tweets",
)
def action(cli_args, enricher, loading_bar):
    for row, tweet in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            tweet_id = tweet

            tweet_parsed = parse_twitter_url(tweet)
            if isinstance(tweet_parsed, TwitterTweet):
                tweet_id = tweet_parsed.id
            elif isinstance(tweet_parsed, (TwitterUser, TwitterList)) or (
                not tweet_parsed and not ID_RE.match(tweet)
            ):
                loading_bar.print("%s is not a tweet id or url." % tweet)
                continue

            try:
                result = get_dates_from_id(tweet_id, locale=cli_args.timezone)

            except (ValueError, TypeError):
                enricher.writerow(row)
                continue

            enricher.writerow(row, list(result))



#!/usr/bin/env python
# =============================================================================
# Minet CLI Endpoint
# =============================================================================
#
# CLI enpoint of the Minet library.
#
import csv
import sys
import ctypes
import importlib
import multiprocessing
import casanova
from casanova.exceptions import MissingColumnError
from contextlib import ExitStack
from encodings import idna  # NOTE: this is necessary for pyinstaller build still...

from minet.cli.constants import DEFAULT_PREBUFFER_BYTES
from minet.loggers import sleepers_logger
from minet.cli.console import console
from minet.cli.utils import (
    die,
    get_rcfile,
    CLIRetryerHandler,
    with_cli_exceptions,
)
from minet.cli.argparse import resolve_arg_dependencies, build_parser, get_subparser
from minet.cli.exceptions import NotResumableError, InvalidArgumentsError, FatalError


@with_cli_exceptions
def run(name, version, commands):
    # Freezing multiprocessing support for pyinstaller etc.
    multiprocessing.freeze_support()

    # Default spawn context for multiprocessing
    multiprocessing.set_start_method("spawn")

    # Increasing max CSV file limit to avoid pesky issues
    csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))

    # Casanova global defaults
    casanova.set_defaults(
        prebuffer_bytes=DEFAULT_PREBUFFER_BYTES,
        strip_null_bytes_on_read=True,
        strip_null_bytes_on_write=True,
    )

    # Adding handlers for sleepers
    sleepers_logger.addHandler(CLIRetryerHandler())

    # Building parser
    parser, subparser_index = build_parser(name, version, commands)

    # Parsing arguments and triggering commands
    cli_args = parser.parse_args()

    # Suppressing console?
    if getattr(cli_args, "silent", False):
        console.quiet = True

    action = subparser_index.get(cli_args.action)

    if action is not None:

        # If subparser is called without any subaction, we print help and exit
        if "subparsers" in action["command"]:
            subdest = action["command"]["subparsers"]["dest"]
            subaction = getattr(cli_args, subdest)

            if subaction is None:
                action["parser"].print_help()
                sys.exit(0)

        # Loading config
        config = get_rcfile(cli_args.rcfile) if hasattr(cli_args, "rcfile") else None

        # Resolving namespace dependencies
        try:
            to_close = resolve_arg_dependencies(cli_args, config)
        except OSError as e:
            parser.error("Could not open output file (-o/--output): %s" % str(e))
        except InvalidArgumentsError as e:
            parser.error(e.message)
        except NotResumableError:
            parser.error(
                "Cannot --resume without knowing where the output will be written (use -o/--output)"
            )

        # Lazy loading module for faster startup
        meta = action["command"]

        if hasattr(cli_args, "subcommand") and cli_args.subcommand:
            meta = action["command"]["subparsers"]["commands"][cli_args.subcommand]

        if "resolve" in meta:
            try:
                meta["resolve"](cli_args)
            except InvalidArgumentsError as e:
                parser.error(e.message)

        fn = meta["package"]

        # Lazy-loading
        if not callable(fn):
            m = importlib.import_module(meta["package"])
            fn = getattr(m, "action")

        with ExitStack() as stack:
            stack.callback(sys.stdout.flush)
            stack.callback(sys.stderr.flush)

            for buffer in to_close:
                stack.callback(buffer.close)

            try:
                fn(cli_args)
            except InvalidArgumentsError as e:
                parser.error(e.message)
            except MissingColumnError as e:
                console.print("Missing column", "[error]{}[/error]!".format(str(e)))
                sys.exit(1)
            except FatalError as e:
                console.vprint(e.message)
                sys.exit(1)

    elif cli_args.action == "help":

        if len(cli_args.subcommand) == 0:
            parser.print_help()
            return

        target = get_subparser(subparser_index, cli_args.subcommand)

        if target is None:
            die('Unknow command "%s"' % " ".join(cli_args.subcommand))
        else:
            target.print_help()

    else:
        parser.print_help()



from minet.cli.argparse import command


def resolve_arguments(cli_args):
    if cli_args.column is None:
        cli_args.column = "filename"


SCRAPE_COMMAND = command(
    "scrape",
    "minet.cli.scrape.scrape",
    title="Minet Scrape Command",
    description="""
        Use multiple processes to scrape data from a batch of HTML files using
        minet scraping DSL documented here:
        https://github.com/medialab/minet/blob/master/docs/cookbook/scraping_dsl.md

        It will output the scraped items as a CSV or NDJSON file.

        Note that this command has been geared towards working in tandem with
        the fetch command. This means the command expects, by default, CSV files
        containing columns like "filename", "http_status", "encoding" etc. as
        you can find in a fetch command CSV report.

        This said, you can of course feed this command any kind of CSV data,
        and use dedicated flags such as --filename-column, --body-column to
        to inform the command about your specific table.

        The comand is also able to work on glob patterns, such as: "downloaded/**/*.html",
        and can also be fed CSV columns containing HTML content directly if
        required.
    """,
    epilog="""
        Builtin scrapers:

        . "canonical": scrape the <link rel="canonical"> tag href if any.
        . "title": scrape the <title> tag if any.
        . "urls": scrape all the relevant <a> tag href urls. Will join them
            with the correct base url if --url-column was given.

        Examples:

        . Scraping a single file on disk:
            $ minet scrape scraper.yml ./path/to/file.html

        . Scraping a `minet fetch` report:
            $ minet scrape scraper.yml -i report.csv -I downloaded > scraped.csv

        . Scraping a single url:
            $ minet fetch "https://lemonde.fr" | minet scrape scraper.yml -i -

        . Indicating a custom `filename` column:
            $ minet scrape scraper.yml -i report.csv -I downloaded --filename-column path > scraped.csv

        . Scraping a CSV colum containing HTML directly:
            $ minet scrape scraper.yml -i report.csv --body-column html > scraped.csv

        . Scraping a bunch of files using a glob pattern:
            $ minet scrape scraper.yml "./content/**/*.html" --glob > scraped.csv

        . Scraping using a CSV file containing glob patterns:
            $ minet scrape scraper.yml pattern -i patterns.csv --glob > scraped.csv

        . Working on a fetch report from stdin (mind the `-`):
            $ minet fetch url file.csv | minet scrape scraper.yml -i - -I downloaded > scraped.csv

        . Yielding items as newline-delimited JSON (jsonl):
            $ minet scrape scraper.yml -i report.csv --format jsonl > scraped.jsonl

        . Using a strainer to optimize performance:
            $ minet scrape links-scraper.yml --strain "a[href]" -i report.csv > links.csv

        . Keeping some columns from input CSV file:
            $ minet scrape scraper.yml -i report.csv -s name,url > scraped.csv

        . Using a builtin scraper:
            $ minet scrape title -i report.csv > titles.csv
    """,
    resolve=resolve_arguments,
    variadic_input={"dummy_column": "filename", "optional": True, "no_help": True},
    arguments=[
        {
            "name": "scraper",
            "help": 'Path to a scraper definition file, or name of a builtin scraper, e.g. "title". See the complete list below.',
        },
        {
            "flags": ["-g", "--glob"],
            "help": "Will interpret given filename as glob patterns to resolve if given.",
            "action": "store_true",
        },
        {
            "flags": ["-I", "--input-dir"],
            "help": "Directory where the HTML files are stored.",
        },
        {
            "flags": ["-p", "--processes"],
            "help": "Number of processes to use. Defaults to roughly half of the available CPUs.",
            "type": int,
        },
        {
            "flags": ["--chunk-size"],
            "help": "Chunk size for multiprocessing.",
            "type": int,
            "default": 1,
        },
        {
            "flag": "--body-column",
            "help": "Name of the CSV column containing html bodies.",
            "default": "body",
        },
        {
            "flag": "--url-column",
            "help": "Name of the CSV column containing the url.",
            "default": "resolved_url",
        },
        {
            "flag": "--error-column",
            "help": "Name of the CSV column containing a fetch error.",
            "default": "fetch_error",
        },
        {
            "flag": "--status-column",
            "help": "Name of the CSV column containing HTTP status.",
            "default": "http_status",
        },
        {
            "flag": "--encoding-column",
            "help": "Name of the CSV column containing file encoding.",
            "default": "encoding",
        },
        {
            "flag": "--mimetype-column",
            "help": "Name of the CSV column containing file mimetype.",
            "default": "mimetype",
        },
        {
            "flag": "--encoding",
            "help": "Name of the default encoding to use. If not given the command will infer it for you.",
        },
        {
            "flags": ["-f", "--format"],
            "help": "Output format.",
            "choices": ["csv", "jsonl", "ndjson"],
            "default": "csv",
        },
        {
            "flag": "--plural-separator",
            "help": "Separator use to join lists of values when serializing to CSV.",
            "default": "|",
        },
        {
            "flag": "--strain",
            "help": "Optional CSS selector used to strain, i.e. only parse matched tags in the parsed html files in order to optimize performance.",
        },
    ],
)



# =============================================================================
# Minet Hyphe Tag CLI Action
# =============================================================================
#
# Logic of the `hyphe tag` action.
#
import casanova

from minet.cli.loading_bar import LoadingBar
from minet.cli.hyphe.utils import with_hyphe_fatal_errors
from minet.cli.exceptions import InvalidArgumentsError
from minet.hyphe import HypheAPIClient
from minet.hyphe.exceptions import HypheRequestFailError


@with_hyphe_fatal_errors
def action(cli_args):
    reader = casanova.reader(cli_args.data, total=cli_args.total)
    headers = reader.headers

    if cli_args.webentity_id_column not in headers:
        raise InvalidArgumentsError(
            "%s column does not exists in given CSV file" % cli_args.webentity_id_column
        )

    tag_pos_map = {}

    for tag_column in cli_args.tag_columns:
        if tag_column not in headers:
            raise InvalidArgumentsError(
                "%s column does not exists in given CSV file" % tag_column
            )

        tag_pos_map[tag_column] = headers[tag_column]

    client = HypheAPIClient(cli_args.url)
    corpus = client.corpus(cli_args.corpus, password=cli_args.password)
    corpus.ensure_is_started()

    with LoadingBar(
        title="Tagging web entitites", unit="webentity", total=reader.total
    ) as loading_bar:
        for row, webentity_id in reader.cells(
            cli_args.webentity_id_column, with_rows=True
        ):
            with loading_bar.step():
                webentity_id = webentity_id.strip()

                if not webentity_id:
                    continue

                try:
                    webentity_id = int(webentity_id)
                except ValueError:
                    loading_bar.print('invalid webentity id "%s"' % webentity_id)
                    continue

                for tag_name, tag_pos in tag_pos_map.items():
                    tag_values = row[tag_pos].strip()

                    if not tag_values:
                        continue

                    tag_values = tag_values.split(cli_args.separator)

                    for tag_value in tag_values:
                        if not tag_value:
                            continue

                        try:
                            corpus.call(
                                "store.add_webentity_tag_value",
                                webentity_id=webentity_id,
                                namespace="USER",
                                category=tag_name,
                                value=tag_value,
                            )
                        except HypheRequestFailError as e:
                            if "could not retrieve WebEntity with id" in str(e):
                                loading_bar.print(
                                    'unkown webentity with id "%s"' % webentity_id
                                )
                                break
                            else:
                                raise



# =============================================================================
# Minet Fetch CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and fetching the urls found
# in the given column. This is done in a respectful multithreaded fashion to
# optimize both running time & memory.
#
from typing import Optional, List, Union, Tuple

import casanova
from casanova import TabularRecord
from dataclasses import dataclass
from datetime import datetime
from ural import is_shortened_url, could_be_html

from minet.executors import RequestResult, HTTPWorkerPayload, HTTPThreadPoolExecutor
from minet.fs import FilenameBuilder, ThreadSafeFilesWriter
from minet.cookies import get_cookie_resolver_from_browser
from minet.web import (
    parse_http_header,
    Response,
    RedirectionStack,
)
from minet.exceptions import InvalidURLError, FilenameFormattingError, HTTPCallbackError
from minet.heuristics import should_spoof_ua_when_resolving
from minet.cli.exceptions import InvalidArgumentsError, FatalError
from minet.cli.reporters import report_error, report_filename_formatting_error
from minet.cli.utils import with_enricher_and_loading_bar, with_ctrl_c_warning


@dataclass
class WorkerAddendum:
    filename: Optional[str] = None
    decoded_contents: Optional[str] = None


@dataclass
class FetchAddendum(TabularRecord):
    resolved_url: Optional[str] = None
    http_status: Optional[int] = None
    datetime_utc: Optional[datetime] = None
    fetch_error: Optional[str] = None
    filename: Optional[str] = None
    mimetype: Optional[str] = None
    encoding: Optional[str] = None

    def infos_from_response(
        self, response: Response, addendum: Optional[WorkerAddendum]
    ) -> None:
        self.resolved_url = response.end_url
        self.http_status = response.status
        self.datetime_utc = response.end_datetime
        self.filename = addendum.filename if addendum else None
        self.encoding = response.encoding
        self.mimetype = response.mimetype


@dataclass
class FetchAddendumWithBody(FetchAddendum):
    body: Optional[str] = None

    def infos_from_response(
        self, response: Response, addendum: Optional[WorkerAddendum]
    ) -> None:
        super().infos_from_response(response, addendum)
        self.body = addendum.decoded_contents if addendum else None


@dataclass
class ResolveAddendum(TabularRecord):
    resolved_url: Optional[str] = None
    http_status: Optional[int] = None
    resolution_error: Optional[str] = None
    redirect_count: Optional[int] = None
    redirect_chain: Optional[List[str]] = None

    def infos_from_stack(self, stack: RedirectionStack):
        last = stack[-1]

        self.resolved_url = last.url
        self.http_status = last.status
        self.redirect_count = len(stack) - 1
        self.redirect_chain = [step.type for step in stack]


def get_style_for_status(status: int) -> str:
    if status < 400:
        return "info"

    return "warning"


def loading_bar_stats_sort_key(item):
    name = item["name"]

    if isinstance(name, int):
        return (0, str(name))

    if name == "filtered":
        return (2, name)

    return (1, name)


def get_headers(cli_args):
    if cli_args.action == "resolve":
        headers = ResolveAddendum
    else:
        headers = FetchAddendum

        if cli_args.contents_in_report:
            headers = FetchAddendumWithBody

    return headers


def get_title(cli_args):
    if cli_args.action == "resolve":
        return "Resolving"

    return "Fetching"


@with_enricher_and_loading_bar(
    headers=get_headers,
    enricher_type="threadsafe",
    index_column="fetch_original_index",
    title=get_title,
    unit="urls",
    stats_sort_key=loading_bar_stats_sort_key,
)
@with_ctrl_c_warning
def action(cli_args, enricher: casanova.ThreadSafeEnricher, loading_bar):
    # Resolving or fetching?
    resolve = cli_args.action == "resolve"

    # HTTP method
    http_method = cli_args.method

    # Cookie grabber
    get_cookie = None
    if cli_args.grab_cookies:
        get_cookie = get_cookie_resolver_from_browser(cli_args.grab_cookies)

    # Global headers
    global_headers = None
    if cli_args.headers:
        global_headers = {}

        for header in cli_args.headers:
            k, v = parse_http_header(header)
            global_headers[k] = v

    if cli_args.column not in enricher.headers:
        raise InvalidArgumentsError(
            'Could not find the "%s" column containing the urls in the given CSV file.'
            % cli_args.column
        )

    url_pos = enricher.headers[cli_args.column]

    filename_pos = None

    if not resolve and cli_args.filename is not None:
        if cli_args.filename not in enricher.headers:
            raise InvalidArgumentsError(
                'Could not find the "%s" column containing the filenames in the given CSV file.'
                % cli_args.filename
            )

        filename_pos = enricher.headers[cli_args.filename]

    only_shortened = getattr(cli_args, "only_shortened", False)
    only_html = getattr(cli_args, "only_html", False)

    def url_key(item) -> Optional[str]:
        url = item[1][url_pos].strip()

        if not url:
            return

        if only_shortened and not is_shortened_url(url):
            return

        if only_html and not could_be_html(url):
            return

        # Url templating
        if cli_args.url_template:
            return cli_args.url_template.format(value=url)

        return url

    def request_args(payload: HTTPWorkerPayload):
        cookie = None

        # Cookie
        if get_cookie:
            cookie = get_cookie(payload.url)

        # Headers
        headers = None

        if global_headers:
            headers = global_headers

        return {
            "method": http_method,
            "cookie": cookie,
            "headers": headers,
            "spoof_ua": resolve and should_spoof_ua_when_resolving(payload.domain),
        }

    # Worker callback internals
    filename_builder = None
    files_writer = None

    if not resolve:
        try:
            filename_builder = FilenameBuilder(
                folder_strategy=cli_args.folder_strategy,
                template=cli_args.filename_template,
            )
        except TypeError:
            raise FatalError(
                [
                    'Invalid "%s" --folder-strategy!' % cli_args.folder_strategy,
                    "Check the list at the end of the command help:",
                    "  $ minet fetch -h",
                ]
            )

        files_writer = ThreadSafeFilesWriter(cli_args.output_dir)

    def worker_callback(
        result: RequestResult[Tuple[int, List[str]], None]
    ) -> Optional[WorkerAddendum]:
        if cli_args.dont_save:
            return

        # NOTE: at this point the callback is only fired on success
        assert result.response

        row = result.item[1]
        response = result.response

        if cli_args.keep_failed_contents and response.status != 200:
            return

        if cli_args.only_html and not response.is_html:
            return

        addendum = WorkerAddendum()

        # First we need to build a filename
        filename_cell = row[filename_pos] if filename_pos is not None else None

        formatter_kwargs = {}

        if cli_args.filename_template and "row" in cli_args.filename_template:
            formatter_kwargs["row"] = enricher.wrap(row)

        assert filename_builder is not None

        filename = filename_builder(
            response.end_url,
            filename=filename_cell,
            ext=response.ext,
            formatter_kwargs=formatter_kwargs,
            compressed=cli_args.compress,
        )

        addendum.filename = filename

        # Decoding the response data?
        data: Union[str, bytes] = response.body

        if response.is_text and (
            cli_args.standardize_encoding or cli_args.contents_in_report
        ):
            data = response.body.decode(response.likely_encoding, errors="replace")

            if cli_args.contents_in_report:
                addendum.decoded_contents = data

        # Writing the file?
        # TODO: specify what should happen when contents are empty (e.g. POST queries)
        if data and not cli_args.contents_in_report:
            assert files_writer is not None

            files_writer.write(filename, data, compress=cli_args.compress)

        return addendum

    common_executor_kwargs = {
        "insecure": cli_args.insecure,
        "max_workers": cli_args.threads,
        "wait": False,
        "daemonic": False,
    }
    common_imap_kwargs = {
        "key": url_key,
        "throttle": cli_args.throttle,
        "domain_parallelism": cli_args.domain_parallelism,
        "max_redirects": cli_args.max_redirects,
    }

    if cli_args.timeout is not None:
        common_executor_kwargs["timeout"] = cli_args.timeout

    # Normal fetch
    if not resolve:

        Addendum = (
            FetchAddendum if not cli_args.contents_in_report else FetchAddendumWithBody
        )

        with HTTPThreadPoolExecutor(**common_executor_kwargs) as executor:
            for result in executor.request(
                enricher,
                request_args=request_args,
                callback=worker_callback,
                **common_imap_kwargs
            ):
                with loading_bar.step():
                    index, row = result.item

                    if not result.url:
                        enricher.writerow(index, row)
                        loading_bar.inc_stat("filtered", style="warning")
                        continue

                    addendum = Addendum()

                    # No error
                    if result.error is None:
                        assert result.response is not None

                        response = result.response
                        status = response.status

                        loading_bar.inc_stat(status, style=get_style_for_status(status))

                        addendum.infos_from_response(response, result.addendum)
                        enricher.writerow(index, row, addendum)

                    # Handling potential errors
                    else:
                        error = result.error

                        if isinstance(error, HTTPCallbackError):
                            error.unwrap(FilenameFormattingError)

                        error_code = report_error(error)

                        loading_bar.inc_stat(error_code, style="error")

                        resolved_url = None

                        if isinstance(error, InvalidURLError):
                            resolved_url = error.url

                        if isinstance(error, FilenameFormattingError):
                            loading_bar.print(report_filename_formatting_error(error))

                        addendum.fetch_error = error_code
                        addendum.resolved_url = resolved_url

                        enricher.writerow(index, row, addendum)

    # Resolve
    else:

        with HTTPThreadPoolExecutor(**common_executor_kwargs) as executor:
            for result in executor.resolve(
                enricher,
                resolve_args=request_args,
                follow_meta_refresh=cli_args.follow_meta_refresh,
                follow_js_relocation=cli_args.follow_js_relocation,
                infer_redirection=cli_args.infer_redirection,
                canonicalize=cli_args.canonicalize,
                **common_imap_kwargs
            ):
                with loading_bar.step():
                    index, row = result.item

                    if not result.url:
                        enricher.writerow(index, row)
                        loading_bar.inc_stat("filtered", style="warning")
                        continue

                    addendum = ResolveAddendum()

                    # No error
                    if result.error is None:
                        assert result.stack is not None

                        # Reporting in output
                        last = result.stack[-1]
                        status = last.status
                        loading_bar.inc_stat(status, style=get_style_for_status(status))

                        addendum.infos_from_stack(result.stack)
                        enricher.writerow(index, row, addendum)

                    # Handling potential errors
                    else:
                        error_code = report_error(result.error)
                        loading_bar.inc_stat(error_code, style="error")

                        addendum.resolution_error = error_code
                        enricher.writerow(index, row, addendum)



# =============================================================================
# Minet Mediacloud Utils
# =============================================================================
#
# Miscellaneous utility function used throughout the mediacloud package.
#
from minet.web import request
from minet.mediacloud.constants import MEDIACLOUD_API_BASE_URL
from minet.mediacloud.exceptions import MediacloudServerError


def get_next_link_id(data):

    if "link_ids" not in data:
        return None

    pagination = data["link_ids"]

    if not pagination.get("next"):
        return None

    return pagination["next"]


def explode_tags(data):
    tags = []
    tag_sets = []
    tags_ids = []
    tag_sets_ids = []

    for tag in data:
        tags.append(tag["tag"])
        tag_sets.append(tag["tag_set"])
        tags_ids.append(tag["tags_id"])
        tag_sets_ids.append(tag["tag_sets_id"])

    return tags, tag_sets, tags_ids, tag_sets_ids


def get_last_processed_stories_id(data):
    if not data:
        return None

    return data[-1]["processed_stories_id"]


def make_simple_call(
    pool_manager, token, route, formatter, raw=False, arg=None, query=None, single=False
):
    url = MEDIACLOUD_API_BASE_URL + route

    if arg is not None:
        url += "/" + str(arg)

    url += "?key=%s" % token

    if query is not None:
        url += "&" + ("&".join("%s=%s" % (str(k), str(v)) for k, v in query.items()))

    response = request(url, pool_manager=pool_manager, known_encoding="utf-8")
    data = response.json()

    if response.status >= 500:
        raise MediacloudServerError(server_error=data.get("error"))

    results = []

    for item in data:
        if not raw:
            item = formatter(item)

        results.append(item)

    if single:
        return results[0]

    return results



# =============================================================================
# Minet HTTP Multithreaded Executors
# =============================================================================
#
# Exposing a specialized quenouille wrapper grabbing various urls from the
# web in a multithreaded fashion.
#
from typing import (
    Optional,
    Iterator,
    Callable,
    TypeVar,
    Generic,
    Iterable,
    Dict,
    Union,
    Any,
)

import urllib3
import threading
from threading import Event
from quenouille import ThreadPoolExecutor
from ural import get_domain_name, ensure_protocol
from tenacity import RetryCallState

from minet.exceptions import CancelledRequestError, HTTPCallbackError
from minet.web import (
    create_pool_manager,
    create_request_retryer,
    request,
    resolve,
    Response,
    RedirectionStack,
    AnyTimeout,
    EXPECTED_WEB_ERRORS,
)
from minet.constants import (
    DEFAULT_DOMAIN_PARALLELISM,
    DEFAULT_IMAP_BUFFER_SIZE,
    DEFAULT_THROTTLE,
    DEFAULT_URLLIB3_TIMEOUT,
    DEFAULT_FETCH_MAX_REDIRECTS,
    DEFAULT_RESOLVE_MAX_REDIRECTS,
)

ItemType = TypeVar("ItemType")
ResultType = TypeVar("ResultType")
AddendumType = TypeVar("AddendumType")

CANCELLED = object()


class HTTPWorkerPayload(Generic[ItemType]):
    __slots__ = ("item", "url", "__has_cached_domain", "__domain")

    item: ItemType
    url: Optional[str]

    __has_cached_domain: bool
    __domain: Optional[str]

    def __init__(self, item: ItemType, url: Optional[str]):
        self.item = item
        self.url = url
        self.__has_cached_domain = False
        self.__domain = None

    @property
    def domain(self) -> Optional[str]:
        if self.__has_cached_domain:
            return self.__domain

        if self.url is not None:
            self.__domain = get_domain_name(self.url)

        self.__has_cached_domain = True

        return self.__domain


ArgsCallbackType = Callable[[HTTPWorkerPayload[ItemType]], Dict]


class RequestResult(Generic[ItemType, AddendumType]):
    __slots__ = ("item", "url", "error", "response", "addendum")

    item: ItemType
    url: Optional[str]
    error: Optional[Exception]
    response: Optional[Response]
    addendum: Optional[AddendumType]

    def __init__(self, item: ItemType, url: Optional[str]):
        self.item = item
        self.url = url
        self.error = None
        self.response = None
        self.addendum = None

    def __repr__(self):
        name = self.__class__.__name__

        if not self.url:
            return "<{name} null!>".format(name=name)

        if not self.response:
            return "<{name} url={url!r} pending!>".format(name=name, url=self.url)

        if self.error:
            return "<{name} url={url!r} error={error}>".format(
                name=name, url=self.url, error=self.error.__class__.__name__
            )

        assert self.response is not None

        return "<{name} url={url!r} status={status!r}>".format(
            name=name, url=self.url, status=self.response.status
        )


class ResolveResult(Generic[ItemType, AddendumType]):
    __slots__ = ("item", "url", "error", "stack", "addendum")

    item: ItemType
    url: Optional[str]
    error: Optional[Exception]
    stack: Optional[RedirectionStack]
    addendum: Optional[AddendumType]

    def __init__(self, item: ItemType, url: Optional[str]):
        self.item = item
        self.url = url
        self.error = None
        self.stack = None
        self.addendum = None

    def __repr__(self):
        name = self.__class__.__name__

        if not self.url:
            return "<{name} null!>".format(name=name)

        if self.error:
            return "<{name} url={url!r} error={error}>".format(
                name=name, url=self.url, error=self.error.__class__.__name__
            )

        assert self.stack is not None

        return "<{name} url={url!r} status={status!r} redirects={redirects!r}>".format(
            name=name,
            url=self.url,
            status=self.stack[-1].status,
            redirects=len(self.stack),
        )


def key_by_domain_name(payload: HTTPWorkerPayload) -> Optional[str]:
    return payload.domain


def payloads_iter(
    iterable: Iterable[ItemType],
    key: Optional[Callable[[ItemType], Optional[str]]] = None,
) -> Iterator[HTTPWorkerPayload[ItemType]]:
    for item in iterable:
        url = item if key is None else key(item)

        if not url:
            yield HTTPWorkerPayload(item=item, url=None)
            continue

        # Url cleanup
        url = ensure_protocol(url.strip())  # type: ignore

        yield HTTPWorkerPayload(item=item, url=url)


class HTTPWorker(Generic[ItemType, AddendumType]):
    def __init__(
        self,
        pool_manager: urllib3.PoolManager,
        cancel_event: Event,
        local_context: threading.local,
        *,
        resolving: bool = False,
        get_args: Optional[ArgsCallbackType[ItemType]] = None,
        max_redirects: int = DEFAULT_FETCH_MAX_REDIRECTS,
        follow_refresh_header: bool = True,
        follow_meta_refresh: bool = False,
        follow_js_relocation: bool = False,
        infer_redirection: bool = False,
        canonicalize: bool = False,
        callback: Optional[
            Callable[[RequestResult[ItemType, AddendumType]], AddendumType]
        ] = None,
    ):
        self.cancel_event = cancel_event
        self.local_context = local_context
        self.get_args = get_args
        self.callback = callback

        self.resolving = resolving
        self.fn = request if not resolving else resolve
        self.Result = RequestResult if not resolving else ResolveResult

        self.default_kwargs = {
            "pool_manager": pool_manager,
            "max_redirects": max_redirects,
            "cancel_event": cancel_event,
            "follow_refresh_header": follow_refresh_header,
            "follow_meta_refresh": follow_meta_refresh,
            "follow_js_relocation": follow_js_relocation,
            "infer_redirection": infer_redirection,
            "canonicalize": canonicalize,
        }

    def __call__(
        self, payload: HTTPWorkerPayload[ItemType]
    ) -> Union[object, RequestResult[ItemType, AddendumType]]:
        item, url = payload.item, payload.url

        result = self.Result(item, url)

        # Noop
        if url is None:
            return result

        kwargs = {}

        if self.cancel_event.is_set():
            return CANCELLED

        if self.get_args is not None:
            # NOTE: given callback must be threadsafe
            kwargs = self.get_args(payload)

        if self.cancel_event.is_set():
            return CANCELLED

        try:
            retryer = getattr(self.local_context, "retryer", None)
            kwargs.update(self.default_kwargs)

            if retryer is not None:
                output = retryer(self.fn, url, **kwargs)
            else:
                output = self.fn(url, **kwargs)

        except CancelledRequestError:
            return CANCELLED

        except EXPECTED_WEB_ERRORS as error:
            result.error = error

        else:
            if self.resolving:
                result.stack = output
            else:
                result.response = output

            if self.callback is not None:
                if self.cancel_event.is_set():
                    return CANCELLED

                try:
                    if retryer is not None:
                        result.addendum = retryer(self.callback, result)
                    else:
                        result.addendum = self.callback(result)
                except Exception as reason:
                    result.error = HTTPCallbackError(reason)

        return result


class HTTPThreadPoolExecutor(ThreadPoolExecutor):
    def __init__(
        self,
        max_workers: Optional[int] = None,
        wait: bool = True,
        daemonic: bool = False,
        insecure: bool = False,
        timeout: AnyTimeout = DEFAULT_URLLIB3_TIMEOUT,
        spoof_tls_ciphers: bool = False,
        proxy: Optional[str] = None,
        retry: bool = False,
        retryer_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        self.cancel_event = Event()
        self.local_context = threading.local()

        if retry:

            def epilog(retry_state: RetryCallState) -> str:
                return retry_state.args[0]

            default_retryer_kwargs = {
                "retry_on_timeout": False,
                "cancel_event": self.cancel_event,
                "max_attempts": 3,
                "epilog": epilog,
            }

            default_retryer_kwargs.update(retryer_kwargs or {})

            def init_local_context():
                self.local_context.retryer = create_request_retryer(
                    **default_retryer_kwargs
                )

            kwargs["initializer"] = init_local_context

        super().__init__(
            max_workers,
            wait=wait,
            daemonic=daemonic,
            **kwargs,
        )
        self.pool_manager = create_pool_manager(
            threads=self.max_workers,
            insecure=insecure,
            timeout=timeout,
            spoof_tls_ciphers=spoof_tls_ciphers,
            proxy=proxy,
        )

    def cancel(self) -> None:
        self.cancel_event.set()

    def shutdown(self, wait=True) -> None:
        self.cancel()
        self.pool_manager.clear()
        return super().shutdown(wait=wait)

    def request(
        self,
        iterator: Iterable[ItemType],
        *,
        ordered: bool = False,
        key: Optional[Callable[[ItemType], Optional[str]]] = None,
        throttle: float = DEFAULT_THROTTLE,
        request_args: Optional[ArgsCallbackType[ItemType]] = None,
        buffer_size: int = DEFAULT_IMAP_BUFFER_SIZE,
        domain_parallelism: int = DEFAULT_DOMAIN_PARALLELISM,
        max_redirects: int = DEFAULT_FETCH_MAX_REDIRECTS,
        callback: Optional[
            Callable[[RequestResult[ItemType, AddendumType]], None]
        ] = None,
    ) -> Iterator[RequestResult[ItemType, AddendumType]]:

        # TODO: validate
        worker = HTTPWorker(
            self.pool_manager,
            self.cancel_event,
            self.local_context,
            get_args=request_args,
            max_redirects=max_redirects,
            callback=callback,
        )

        method = super().imap if ordered else super().imap_unordered

        imap_unordered = method(
            payloads_iter(iterator, key=key),
            worker,
            key=key_by_domain_name,
            parallelism=domain_parallelism,
            buffer_size=buffer_size,
            throttle=throttle,
        )

        return (item for item in imap_unordered if item is not CANCELLED)  # type: ignore

    def resolve(
        self,
        iterator: Iterable[ItemType],
        *,
        ordered: bool = False,
        key: Optional[Callable[[ItemType], Optional[str]]] = None,
        throttle: float = DEFAULT_THROTTLE,
        resolve_args: Optional[ArgsCallbackType[ItemType]] = None,
        buffer_size: int = DEFAULT_IMAP_BUFFER_SIZE,
        domain_parallelism: int = DEFAULT_DOMAIN_PARALLELISM,
        max_redirects: int = DEFAULT_RESOLVE_MAX_REDIRECTS,
        follow_refresh_header: bool = True,
        follow_meta_refresh: bool = False,
        follow_js_relocation: bool = False,
        infer_redirection: bool = False,
        canonicalize: bool = False,
        callback: Optional[
            Callable[[ResolveResult[ItemType, AddendumType]], None]
        ] = None,
    ) -> Iterator[ResolveResult[ItemType, AddendumType]]:

        # TODO: validate
        worker = HTTPWorker(
            self.pool_manager,
            self.cancel_event,
            self.local_context,
            resolving=True,
            get_args=resolve_args,
            max_redirects=max_redirects,
            follow_refresh_header=follow_refresh_header,
            follow_meta_refresh=follow_meta_refresh,
            follow_js_relocation=follow_js_relocation,
            infer_redirection=infer_redirection,
            canonicalize=canonicalize,
            callback=callback,
        )

        method = super().imap if ordered else super().imap_unordered

        imap_unordered = method(
            payloads_iter(iterator, key=key),
            worker,
            key=key_by_domain_name,
            parallelism=domain_parallelism,
            buffer_size=buffer_size,
            throttle=throttle,
        )

        return (item for item in imap_unordered if item is not CANCELLED)  # type: ignore



from typing import Optional

import browser_cookie3
from urllib.request import Request
from http.cookiejar import CookieJar
from http.cookies import SimpleCookie

from minet.exceptions import UnknownBrowserError, CookieGrabbingError
from minet.constants import COOKIE_BROWSERS


class CookieResolver(object):
    def __init__(self, jar: CookieJar):
        self.jar = jar

    def __call__(self, url: str) -> Optional[str]:
        req = Request(url)
        self.jar.add_cookie_header(req)

        return req.get_header("Cookie") or None


def get_cookie_jar_from_browser(browser="firefox") -> CookieJar:
    fn = getattr(browser_cookie3, browser, None)

    if fn is None:
        raise UnknownBrowserError(browser)

    try:
        return fn()
    except browser_cookie3.BrowserCookieError:
        raise CookieGrabbingError(browser)


def get_cookie_resolver_from_browser(browser="firefox") -> CookieResolver:
    return CookieResolver(get_cookie_jar_from_browser(browser))


def coerce_cookie_for_url_from_browser(target: str, url: str) -> Optional[str]:
    if target in COOKIE_BROWSERS:
        get = get_cookie_resolver_from_browser(target)

        return get(url)

    return target.strip()


def get_cookie_morsel_value(cookie, key):
    parsed = SimpleCookie()
    parsed.load(cookie)
    return parsed[key].value


def dict_to_cookie_string(d):
    return "; ".join("%s=%s" % r for r in d.items())



# =============================================================================
# Minet Crawl CLI Action
# =============================================================================
#
# Logic of the crawl action.
#
from typing import List, Any, Optional, Union, TextIO, Tuple

import os
import casanova
from os.path import join, isfile, isdir, dirname
from ebbe.decorators import with_defer

from minet.cli.exceptions import FatalError
from minet.cli.console import console
from minet.scrape import Scraper
from minet.scrape.exceptions import InvalidScraperError
from minet.crawl import Crawler, CrawlResult, CrawlerState, DefinitionSpiderOutput
from minet.cli.reporters import report_error, report_scraper_validation_errors
from minet.cli.loading_bar import LoadingBar
from minet.cli.utils import with_loading_bar, with_ctrl_c_warning

JOBS_HEADERS = [
    "spider",
    "depth",
    "url",
    "error",
    "status",
    "encoding",
    "degree",
    "scraped",
]

STATUS_TO_STYLE = {
    "acked": "success_background",
    "ready": "info_background",
    "unack": "warning_background",
    "ack_failed": "error_background",
}


def format_result_for_csv(
    result: CrawlResult[Any, DefinitionSpiderOutput], count: Optional[int] = None
) -> List[Optional[Union[str, int]]]:
    if result.error is not None:
        return [
            result.job.spider,
            result.job.depth,
            result.job.url,
            report_error(result.error),
            "",
            "",
            "",
            "",
            "",
        ]

    response = result.response

    assert response is not None

    return [
        result.job.spider,
        result.job.depth,
        result.job.url,
        "",
        response.status,
        response.encoding,
        result.degree,
        count,
    ]


def open_report(
    path: str, headers: List[str], resume: bool = False
) -> Tuple[TextIO, casanova.Writer]:
    flag = "w"

    if resume and isfile(path):
        flag = "a"

    os.makedirs(dirname(path), exist_ok=True)

    f = open(path, flag, encoding="utf-8")
    writer = casanova.writer(f)

    if flag == "w":
        writer.writerow(headers)

    return f, writer


class ScraperReporter(object):
    def __init__(self, path: str, scraper: Scraper, resume=False):
        if scraper.fieldnames is None:
            raise NotImplementedError("Scraper fieldnames could not be inferred.")

        f, writer = open_report(path, scraper.fieldnames, resume)

        self.fieldnames = scraper.fieldnames
        self.file = f
        self.writer = writer

    def write(self, items) -> int:
        count = 0

        # TODO: maybe abstract this once step above
        if not isinstance(items, list):
            items = [items]

        for item in items:
            count += 1

            if not isinstance(item, dict):
                self.writer.writerow([item])
                continue

            row = [item.get(k, "") for k in self.fieldnames]
            self.writer.writerow(row)

        return count

    def flush(self):
        self.file.flush()

    def close(self):
        self.file.close()


class ScraperReporterPool(object):
    SINGULAR = "$SINGULAR$"

    def __init__(self, crawler: Crawler, output_dir: str, resume=False):
        self.reporters = {}

        if crawler.singular:
            spider = crawler.get_spider()

            self.reporters["default"] = {}

            if spider.scraper is not None:
                path = join(output_dir, "scraped.csv")
                reporter = ScraperReporter(path, spider.scraper, resume)
                self.reporters["default"][ScraperReporterPool.SINGULAR] = reporter

            for name, scraper in spider.scrapers.items():
                path = join(output_dir, "scraped", "%s.csv" % name)

                reporter = ScraperReporter(path, scraper, resume)
                self.reporters["default"][name] = reporter
        else:
            for spider_name, spider in crawler.spiders.items():
                self.reporters[spider_name] = {}

                if spider.scraper is not None:
                    path = join(output_dir, "scraped", spider_name, "scraped.csv")
                    reporter = ScraperReporter(path, spider.scraper, resume)
                    self.reporters[spider_name][ScraperReporterPool.SINGULAR] = reporter

                for name, scraper in spider.scrapers.items():
                    path = join(output_dir, "scraped", spider_name, "%s.csv" % name)

                    reporter = ScraperReporter(path, scraper, resume)
                    self.reporters[spider_name][name] = reporter

    def write(self, spider_name: Optional[str], scraped: DefinitionSpiderOutput) -> int:
        count = 0

        if spider_name is None:
            spider_name = "default"

        reporter = self.reporters[spider_name]

        if scraped.default is not None:
            count += reporter[ScraperReporterPool.SINGULAR].write(scraped.default)

        for name, items in scraped.named.items():
            count += reporter[name].write(items)

        return count

    def __iter__(self):
        for spider_reporters in self.reporters.values():
            for reporter in spider_reporters.values():
                yield reporter

    def flush(self) -> None:
        for reporter in self:
            reporter.flush()

    def close(self) -> None:
        for reporter in self:
            reporter.close()


@with_defer()
@with_loading_bar(
    title="Crawling",
    unit="pages",
    stats=[
        {"name": "queued", "style": "info"},
        {"name": "doing", "style": "warning"},
        {"name": "done", "style": "success"},
    ],
)
@with_ctrl_c_warning
def action(cli_args, defer, loading_bar: LoadingBar):

    if cli_args.dump_queue and not isdir(cli_args.output_dir):
        loading_bar.erase()
        raise FatalError("Cannot dump crawl not started yet!")

    # Loading crawler definition
    queue_path = join(cli_args.output_dir, "queue")

    # Scaffolding output directory
    os.makedirs(cli_args.output_dir, exist_ok=True)

    jobs_output_path = join(cli_args.output_dir, "jobs.csv")
    jobs_output, jobs_writer = open_report(
        jobs_output_path, JOBS_HEADERS, resume=cli_args.resume
    )
    defer(jobs_output.close)

    # Creating crawler
    try:
        crawler = Crawler.from_definition(
            cli_args.crawler,
            throttle=cli_args.throttle,
            queue_path=queue_path,
            resume=cli_args.resume,
            wait=False,
            daemonic=False,
        )
    except InvalidScraperError as error:
        raise FatalError(
            [
                "Your scraper is invalid! You need to fix the following errors:\n",
                report_scraper_validation_errors(error.validation_errors),
            ]
        )

    if cli_args.dump_queue:
        loading_bar.erase()
        dump = crawler.dump_queue()
        for (status, job) in dump:
            console.print(
                "[{style}]{status}[/{style}] depth=[warning]{depth}[/warning] {url}".format(
                    style=STATUS_TO_STYLE.get(status, "log.time"),
                    status=status,
                    depth=job.depth,
                    url=job.url,
                ),
                no_wrap=True,
                overflow="ellipsis",
            )

        console.print("Total:", "[success]{}".format(len(dump)))
        crawler.stop()
        return

    if crawler.finished:
        loading_bar.erase()
        crawler.stop()
        raise FatalError("[error]Crawler has already finished!")

    if crawler.resuming:
        loading_bar.print("[log.time]Will now resume…")

    with crawler:

        # Reporter pool
        reporter_pool = ScraperReporterPool(
            crawler, cli_args.output_dir, resume=cli_args.resume
        )
        defer(reporter_pool.close)

        def on_state_update(state: CrawlerState):
            loading_bar.set_total(state.total)
            loading_bar.set_stat("queued", state.jobs_queued)
            loading_bar.set_stat("doing", state.jobs_doing)
            loading_bar.set_stat("done", state.jobs_done)

        crawler.state.set_listener(on_state_update)

        # Running crawler
        for result in crawler:
            with loading_bar.step():
                if result.error is not None:
                    loading_bar.inc_stat(report_error(result.error), style="error")
                    jobs_writer.writerow(format_result_for_csv(result))
                    continue

                count = reporter_pool.write(result.job.spider, result.data)
                jobs_writer.writerow(format_result_for_csv(result, count=count))
                loading_bar.inc_stat("scraped", count=count, style="success")

                # Flushing to avoid sync issues as well as possible
                jobs_output.flush()
                reporter_pool.flush()



# =============================================================================
# Minet Mediacloud Constants
# =============================================================================
#
# Bunch of mediacloud-related constants.
#
from urllib3 import Timeout


MEDIACLOUD_API_BASE_URL = "https://api.mediacloud.org/api/v2"
MEDIACLOUD_DEFAULT_TIMEOUT = Timeout(connect=30, read=60 * 5)
MEDIACLOUD_DEFAULT_BATCH = 250

MEDIACLOUD_TOPIC_STORIES_CSV_HEADERS = [
    "guid",
    "stories_id",
    "title",
    "url",
    "language",
    "media_id",
    "media_name",
    "collect_date",
    "publish_date",
    "date_is_reliable",
    "facebook_share_count",
    "full_text_rss",
    "inlink_count",
    "outlink_count",
    "media_inlink_count",
    "post_count",
    "snapshots_id",
    "timespans_id",
    "next_link_id",
]

MEDIACLOUD_STORIES_CSV_HEADER = [
    "guid",
    "stories_id",
    "processed_stories_id",
    "title",
    "url",
    "language",
    "collect_date",
    "publish_date",
    "media_id",
    "media_name",
    "media_url",
    "tags",
    "tag_sets",
    "tags_ids",
    "tag_sets_ids",
]

MEDIACLOUD_MEDIA_CSV_HEADER = [
    "media_id",
    "media_name",
    "media_url",
    "is_healthy",
    "is_monitored",
    "public_notes",
    "num_stories_90",
    "num_sentences_90",
    "start_date",
    "tags",
    "tag_sets",
    "tags_ids",
    "tag_sets_ids",
]

MEDIACLOUD_FEED_CSV_HEADER = [
    "name",
    "url",
    "feeds_id",
    "type",
    "media_id",
    "active",
    "last_attempted_download_time",
    "last_new_story_time",
    "last_successful_download_time",
]



from typing import TypeVar, Mapping, Union, Generic, Optional

from ural import get_domain_name, ensure_protocol

from minet.web import Response

CrawlJobDataType = TypeVar("CrawlJobDataType", bound=Mapping)
CrawlResultDataType = TypeVar("CrawlResultDataType")


class CrawlJob(Generic[CrawlJobDataType]):
    __slots__ = (
        "url",
        "depth",
        "spider",
        "data",
        "attempts",
        "__has_cached_domain",
        "__domain",
    )

    url: str
    depth: int
    spider: Optional[str]
    data: Optional[CrawlJobDataType]
    attempts: int

    # TODO: we should add headers, cookies and such here in the future

    __has_cached_domain: bool
    __domain: Optional[str]

    def __init__(
        self,
        url: str,
        depth: Optional[int] = None,
        spider: Optional[str] = None,
        data: Optional[CrawlJobDataType] = None,
    ):
        self.url = ensure_protocol(url).strip()
        self.depth = depth if depth is not None else 0
        self.spider = spider
        self.data = data
        self.attempts = 0

        self.__has_cached_domain = False
        self.__domain = None

    def __getstate__(self):
        return (self.url, self.depth, self.spider, self.data, self.attempts)

    def __setstate__(self, state):
        self.url = state[0]
        self.depth = state[1]
        self.spider = state[2]
        self.data = state[3]
        self.attempts = state[4]

        self.__has_cached_domain = False
        self.__domain = None

    @property
    def domain(self) -> Optional[str]:
        if self.__has_cached_domain:
            return self.__domain

        if self.url is not None:
            self.__domain = get_domain_name(self.url)

        self.__has_cached_domain = True

        return self.__domain

    def __repr__(self):
        class_name = self.__class__.__name__

        return (
            "<{class_name} depth={depth!r} url={url!r} spider={spider!r} attempts={attempts!r}>"
        ).format(
            class_name=class_name,
            url=self.url,
            depth=self.depth,
            spider=self.spider,
            attempts=self.attempts,
        )


UrlOrCrawlJob = Union[str, CrawlJob[CrawlJobDataType]]


class CrawlResult(Generic[CrawlJobDataType, CrawlResultDataType]):
    __slots__ = ("job", "data", "error", "response", "degree")

    job: CrawlJob[CrawlJobDataType]
    data: Optional[CrawlResultDataType]
    error: Optional[Exception]
    response: Optional[Response]
    degree: int

    def __init__(self, job: CrawlJob[CrawlJobDataType]):
        self.job = job
        self.data = None
        self.error = None
        self.response = None
        self.degree = 0


class ErroredCrawlResult(CrawlResult[CrawlJobDataType, None]):
    job: CrawlJob[CrawlJobDataType]
    data: None
    error: Exception
    response: Optional[Response]
    degree: int

    def __init__(
        self,
        job: CrawlJob[CrawlJobDataType],
        error: Exception,
        response: Optional[Response] = None,
    ):
        self.job = job
        self.data = None
        self.error = error
        self.response = response
        self.degree = 0

    def __repr__(self):
        name = self.__class__.__name__

        return "<{name} url={url!r} error={error}>".format(
            name=name, url=self.job.url, error=self.error.__class__.__name__
        )


class SuccessfulCrawlResult(CrawlResult[CrawlJobDataType, CrawlResultDataType]):
    job: CrawlJob[CrawlJobDataType]
    data: CrawlResultDataType
    error: None
    response: Response
    degree: int

    def __init__(
        self,
        job: CrawlJob[CrawlJobDataType],
        response: Response,
        data: CrawlResultDataType,
        degree: int,
    ):
        self.job = job
        self.data = data
        self.error = None
        self.response = response
        self.degree = degree

    def __repr__(self):
        name = self.__class__.__name__

        return "<{name} url={url!r} status={status!r} degree={degree!r}>".format(
            name=name, url=self.job.url, status=self.response.status, degree=self.degree
        )


SomeCrawlResult = Union[
    ErroredCrawlResult[CrawlJobDataType],
    SuccessfulCrawlResult[CrawlJobDataType, CrawlResultDataType],
]



# =============================================================================
# Minet BuzzSumo Domain CLI Action
# =============================================================================
#
# Logic of the `bz domain` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.buzzsumo.utils import with_buzzsumo_fatal_errors
from minet.buzzsumo import BuzzSumoAPIClient
from minet.buzzsumo.constants import ARTICLES_CSV_HEADERS


@with_buzzsumo_fatal_errors
@with_enricher_and_loading_bar(
    title="Retrieving articles",
    unit="articles",
    headers=ARTICLES_CSV_HEADERS,
    nested=True,
)
def action(cli_args, enricher, loading_bar):
    client = BuzzSumoAPIClient(cli_args.token)

    for row, domain_name in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(domain_name):

            for article in client.domain_articles(
                domain_name, cli_args.begin_date, cli_args.end_date
            ):
                loading_bar.nested_advance()
                enricher.writerow(row, article.as_csv_row())



# =============================================================================
# Minet Facebook Exceptions
# =============================================================================
#
from minet.exceptions import MinetError


class FacebookError(MinetError):
    pass


class FacebookInvalidCookieError(FacebookError):
    def __init__(self, msg=None, target=None):
        super().__init__(msg)
        self.target = target


class FacebookInvalidTargetError(FacebookError):
    pass


class FacebookNotPostError(FacebookError):
    pass


class FacebookWatchError(FacebookError):
    pass



# =============================================================================
# Minet Scraper Constants
# =============================================================================
#
# Constants used throughout the minet.scrape package.
#
EXTRACTOR_NAMES = ["text", "display_text", "html", "inner_html", "outer_html"]
SELECT_ALIASES = ["sel", "$"]
ITERATOR_ALIASES = ["iterator", "$$"]
PLURAL_MODIFIERS = ["filter", "filter_eval", "uniq"]
BURROWING_KEYS = ITERATOR_ALIASES + ["item", "fields"]
LEAF_KEYS = ["attr", "extract", "get_context", "eval", "default"]
KNOWN_KEYS = (
    SELECT_ALIASES
    + PLURAL_MODIFIERS
    + BURROWING_KEYS
    + LEAF_KEYS
    + ["sel_eval", "iterator_eval"]
)

BLOCK_ELEMENTS = {
    "article",
    "aside",
    "blockquote",
    "body",
    "br",
    "button",
    "canvas",
    "caption",
    "col",
    "colgroup",
    "dd",
    "div",
    "dl",
    "dt",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "header",
    "hgroup",
    "hr",
    "li",
    "map",
    "object",
    "ol",
    "output",
    "p",
    "pre",
    "progress",
    "section",
    "table",
    "tbody",
    "textarea",
    "tfoot",
    "thead",
    "tr",
    "ul",
    "video",
}

CONTENT_BLOCK_ELEMENTS = {
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "p",
    "ul",
    "ol",
    "pre",
    "blockquote",
}



# =============================================================================
# Minet Twitter Scrape CLI Action
# =============================================================================
#
# Logic of the `tw scrape` action.
#
from twitwi.constants import TWEET_FIELDS, USER_FIELDS
from twitwi import format_tweet_as_csv_row, format_user_as_csv_row

from minet.utils import PseudoFStringFormatter
from minet.cli.utils import with_enricher_and_loading_bar
from minet.twitter import TwitterAPIScraper
from minet.twitter.exceptions import (
    TwitterPublicAPIQueryTooLongError,
)
from minet.twitter.constants import ADDITIONAL_TWEET_FIELDS

CUSTOM_FORMATTER = PseudoFStringFormatter()


def format_meta_row(meta):
    if meta is None:
        return [""] * len(ADDITIONAL_TWEET_FIELDS)

    return [
        meta.get("intervention_type"),
        meta.get("intervention_text"),
        meta.get("intervention_url"),
    ]


def get_headers(cli_args):
    return (
        (TWEET_FIELDS + ADDITIONAL_TWEET_FIELDS)
        if cli_args.items == "tweets"
        else USER_FIELDS
    )


@with_enricher_and_loading_bar(
    headers=get_headers,
    title="Scraping",
    unit="queries",
    nested=True,
    sub_unit=lambda cli_args: cli_args.items,
)
def action(cli_args, enricher, loading_bar):
    scraper = TwitterAPIScraper(cli_args.cookie)

    for row, query in enricher.cells(cli_args.column, with_rows=True):

        # Templating?
        if cli_args.query_template is not None:
            query = CUSTOM_FORMATTER.format(cli_args.query_template, value=query)

        with loading_bar.step(query):
            if cli_args.items == "tweets":
                iterator = scraper.search_tweets(
                    query,
                    locale=cli_args.timezone,
                    limit=cli_args.limit,
                    include_referenced_tweets=cli_args.include_refs,
                    with_meta=True,
                )
            else:
                iterator = scraper.search_users(
                    query, locale=cli_args.timezone, limit=cli_args.limit
                )

            try:
                for data in iterator:
                    if cli_args.items == "tweets":
                        tweet, meta = data
                        tweet_row = format_tweet_as_csv_row(tweet)
                        addendum = tweet_row + format_meta_row(meta)
                    else:
                        addendum = format_user_as_csv_row(data)

                    enricher.writerow(row, addendum)
                    loading_bar.nested_advance()

            except TwitterPublicAPIQueryTooLongError:
                loading_bar.error("Query too long [dim]%s[/dim]" % query)



# =============================================================================
# Minet Cookies CLI Action
# =============================================================================
#
# Logic of the cookies action.
#
import csv
import time
import browser_cookie3
from http.cookies import SimpleCookie

from minet.cookies import CookieResolver
from minet.cli.exceptions import FatalError


# Taken from: https://github.com/python/cpython/blob/3.9/Lib/http/cookiejar.py
def write_jar_as_text_mozilla(jar, f, ignore_discard=False, ignore_expires=False):
    now = time.time()
    for cookie in jar:
        if not ignore_discard and cookie.discard:
            continue
        if not ignore_expires and cookie.is_expired(now):
            continue
        if cookie.secure:
            secure = "TRUE"
        else:
            secure = "FALSE"
        if cookie.domain.startswith("."):
            initial_dot = "TRUE"
        else:
            initial_dot = "FALSE"
        if cookie.expires is not None:
            expires = str(cookie.expires)
        else:
            expires = ""
        if cookie.value is None:
            # cookies.txt regards 'Set-Cookie: foo' as a cookie
            # with no name, whereas http.cookiejar regards it as a
            # cookie with no value.
            name = ""
            value = cookie.name
        else:
            name = cookie.name
            value = cookie.value
        f.write(
            "\t".join(
                [cookie.domain, initial_dot, cookie.path, secure, expires, name, value]
            )
            + "\n"
        )


COOKIE_CSV_HEADER = [
    "domain",
    "name",
    "value",
    "path",
    "secure",
    "expires",
    "is_expired",
]

MORSEL_CSV_HEADER = ["key", "value"]


def format_cookie_for_csv(cookie):
    return [
        cookie.domain,
        cookie.name,
        cookie.value,
        cookie.path,
        "true" if cookie.secure else "false",
        cookie.expires,
        "true" if cookie.is_expired else "false",
    ]


def format_morsel_for_csv(morsel):
    return [morsel.key, morsel.value]


def action(cli_args):
    if cli_args.csv:
        output_writer = csv.writer(cli_args.output)

    try:
        jar = getattr(browser_cookie3, cli_args.browser)()
    except browser_cookie3.BrowserCookieError:
        raise FatalError("Could not extract cookies from %s!" % cli_args.browser)

    if cli_args.url is not None:
        resolver = CookieResolver(jar)

        cookie = resolver(cli_args.url)

        if cookie is not None:

            if cli_args.csv:
                output_writer.writerow(MORSEL_CSV_HEADER)

                parsed = SimpleCookie(cookie)

                for morsel in parsed.values():
                    output_writer.writerow(format_morsel_for_csv(morsel))
            else:
                print(cookie, file=cli_args.output)
        else:
            raise FatalError(
                "Could not find relevant cookie for %s in %s!"
                % (cli_args.url, cli_args.browser)
            )
    else:
        if cli_args.csv:
            output_writer.writerow(COOKIE_CSV_HEADER)

            for cookie in jar:
                output_writer.writerow(format_cookie_for_csv(cookie))
        else:
            write_jar_as_text_mozilla(jar, cli_args.output)



# =============================================================================
# Minet Instagram CLI Action
# =============================================================================
#
# Logic of the `insta` action.
#
from minet.cli.argparse import command, subcommand, ConfigAction

INSTAGRAM_COMMENTS_SUBCOMMAND = subcommand(
    "comments",
    "minet.cli.instagram.comments",
    title="Instagram Comments Command",
    description="""
        Scrape Instagram comments with a given post url, post shortcode or post id.

        This requires to be logged in to an Instagram account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.
    """,
    epilog="""
        example:

        . Searching comments from the post https://www.instagram.com/p/CpA46rmU26Y/:
            $ minet instagram comments https://www.instagram.com/p/CpA46rmU26Y/ > comments.csv
    """,
    variadic_input={
        "dummy_column": "post",
        "item_label": "post url, post shortcode or post id",
        "item_label_plural": "post urls, post shortcodes or post ids",
    },
    select=True,
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of comments to retrieve per post.",
            "type": int,
        }
    ],
)

INSTAGRAM_HASHTAG_SUBCOMMAND = subcommand(
    "hashtag",
    "minet.cli.instagram.hashtag",
    title="Instagram hashtag",
    description="""
        Scrape Instagram posts with a given hashtag.

        This requires to be logged in to an Instagram account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        display_url is not the media url, but a thumbnail of the post.
        There is no way with this command to get the media urls.
    """,
    epilog="""
        example:

        . Searching posts with the hashtag paris:
            $ minet instagram hashtag paris > paris_posts.csv
    """,
    variadic_input={"dummy_column": "hashtag"},
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of posts to retrieve per hashtag.",
            "type": int,
        },
    ],
)

INSTAGRAM_USER_FOLLOWERS_SUBCOMMAND = subcommand(
    "user-followers",
    "minet.cli.instagram.user_followers",
    title="Instagram User Followers Command",
    description="""
        Scrape Instagram followers with a given username, user url or user id.
        On verified accounts, you may be unable to get all of them.

        This requires to be logged in to an Instagram account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        Beware, instagram only provides temporary links, not permalinks,
        for profile picture urls retrieved as the "profile_pic_url" in
        the result. Be sure to download them fast if you need them (you can
        use the `minet fetch` command for that, and won't need to use cookies).

        If a username is a number without '@' at the begining, it will be
        considered as an id.
    """,
    epilog="""
        example:

        . Searching followers with the username banksrepeta:
            $ minet instagram user-followers banksrepeta > banksrepeta_followers.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "username, user url or user id",
        "item_label_plural": "usernames, user urls or user ids",
    },
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of followers to retrieve per user.",
            "type": int,
        },
    ],
)

INSTAGRAM_POST_INFOS_SUBCOMMAND = subcommand(
    "post-infos",
    "minet.cli.instagram.post_infos",
    title="Instagram post-infos",
    description="""
        Scrape Instagram infos with a given post url, post shortcode or post id.

        This requires to be logged in to an Instagram account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        Beware, instagram only provides temporary links, not permalinks,
        for profile picture urls retrieved as the "profile_pic_url_hd" in
        the result. Be sure to download them fast if you need them (you can
        use the `minet fetch` command for that, and won't need to use cookies).
    """,
    epilog="""
        example:

        . Searching infos for the post https://www.instagram.com/p/CpA46rmU26Y/:
            $ minet instagram post-infos https://www.instagram.com/p/CpA46rmU26Y/ > post_infos.csv
    """,
    variadic_input={
        "dummy_column": "post",
        "item_label": "post url, post shortcode or post id",
        "item_label_plural": "post urls, post shortcodes or post ids",
    },
    select=True,
)

INSTAGRAM_USER_FOLLOWING_SUBCOMMAND = subcommand(
    "user-following",
    "minet.cli.instagram.user_following",
    title="Instagram User Following Command",
    description="""
        Scrape Instagram accounts followed with a given username, user url or user id.

        This requires to be logged in to an Instagram account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        Beware, instagram only provides temporary links, not permalinks,
        for profile picture urls retrieved as the "profile_pic_url" in
        the result. Be sure to download them fast if you need them (you can
        use the `minet fetch` command for that, and won't need to use cookies).

        If a username is a number without '@' at the begining, it will be
        considered as an id.
    """,
    epilog="""
        example:

        . Searching accounts followed with the username paramountplus:
            $ minet instagram user-following paramountplus > paramountplus_following.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "username, user url or user id",
        "item_label_plural": "usernames, user urls or user ids",
    },
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of accounts to retrieve per user.",
            "type": int,
        },
    ],
)

INSTAGRAM_USER_INFOS_SUBCOMMAND = subcommand(
    "user-infos",
    "minet.cli.instagram.user_infos",
    title="Instagram user-infos",
    description="""
        Scrape Instagram infos with a given username, user url or user id.

        This requires to be logged in to an Instagram account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        Beware, instagram only provides temporary links, not permalinks,
        for profile picture urls retrieved as the "profile_pic_url_hd" in
        the result. Be sure to download them fast if you need them (you can
        use the `minet fetch` command for that, and won't need to use cookies).

        If a username is a number without '@' at the begining, it will be
        considered as an id.
    """,
    epilog="""
        example:

        . Searching infos with the username banksrepeta:
            $ minet instagram user-infos banksrepeta > banksrepeta_infos.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "username, user url or user id",
        "item_label_plural": "usernames, user urls or user ids",
    },
)

INSTAGRAM_USER_POSTS_SUBCOMMAND = subcommand(
    "user-posts",
    "minet.cli.instagram.user_posts",
    title="Instagram User Posts Command",
    description="""
        Scrape Instagram posts with a given username, user url or user id.

        This requires to be logged in to an Instagram account, so
        by default this command will attempt to grab the relevant
        authentication cookies from a local Firefox browser.

        If you want to grab cookies from another browser or want
        to directly pass the cookie as a string, check out the
        -c/--cookie flag.

        The urls in the medias_url column have a limited life time.
        It is not the case for the url in main_thumbnail_url, which
        corresponds to the first image (the video cover if the first
        media is a video). Be sure to download them fast if you need
        them (you can use the `minet fetch` command for that, and
        won't need to use cookies).

        If a username is a number without '@' at the begining, it will be
        considered as an id.
    """,
    epilog="""
        example:

        . Searching posts from the account paramountplus:
            $ minet instagram user-posts paramountplus > paramountplus_posts.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "username, user url or user id",
        "item_label_plural": "usernames, user urls or user ids",
    },
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of posts to retrieve per user.",
            "type": int,
        }
    ],
)

INSTAGRAM_COMMAND = command(
    "instagram",
    "minet.cli.instagram",
    aliases=["insta"],
    title="Minet Instagram Command",
    description="""
        Gather data from Instagram.
    """,
    common_arguments=[
        {
            "flags": ["-c", "--cookie"],
            "help": 'Authenticated cookie to use or browser from which to extract it (supports "firefox", "chrome", "chromium", "opera" and "edge").',
            "default": "firefox",
            "rc_key": ["instagram", "cookie"],
            "action": ConfigAction,
        }
    ],
    subcommands=[
        INSTAGRAM_COMMENTS_SUBCOMMAND,
        INSTAGRAM_HASHTAG_SUBCOMMAND,
        INSTAGRAM_POST_INFOS_SUBCOMMAND,
        INSTAGRAM_USER_FOLLOWERS_SUBCOMMAND,
        INSTAGRAM_USER_FOLLOWING_SUBCOMMAND,
        INSTAGRAM_USER_INFOS_SUBCOMMAND,
        INSTAGRAM_USER_POSTS_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet Facebook Url Likes CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and retrieving approximate
# likes count by scraping Facebook's like button plugin.
#
from typing import Optional

import re
from urllib.parse import quote
from ural import is_url

from minet.utils import rate_limited
from minet.web import request
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.exceptions import FatalError

REPORT_HEADERS = ["approx_likes", "approx_likes_int"]
NUMBER_RE = re.compile(rb">(\d+\.?\d*[KM]?)<", re.I)


def forge_url(url):
    return (
        "https://www.facebook.com/plugins/share_button.php?href=%s&layout=button_count"
        % quote(url)
    )


@rate_limited(5)
def make_request(url: str) -> Optional[bytes]:
    response = request(forge_url(url), headers={"Accept-Language": "en"})

    if response.status == 404:
        return None

    return response.body


def scrape(data: bytes):

    match = NUMBER_RE.search(data)

    if match is None:
        return ["", ""]

    approx_likes = match.group(1).decode()
    approx_likes_int = approx_likes

    if "K" in approx_likes:
        approx_likes_int = str(int(float(approx_likes[:-1]) * 10**3))

    elif "M" in approx_likes:
        approx_likes_int = str(int(float(approx_likes[:-1]) * 10**6))

    return [approx_likes, approx_likes_int]


@with_enricher_and_loading_bar(
    headers=REPORT_HEADERS, title="Scraping likes", unit="urls"
)
def action(cli_args, enricher, loading_bar):
    for row, url in enricher.cells(cli_args.column, with_rows=True):

        with loading_bar.step():
            url = url.strip()

            if not url or not is_url(url, require_protocol=False):
                enricher.writerow(row)
                continue

            try:
                html = make_request(url)
            except Exception:
                raise FatalError(
                    "An error occurred while fetching like button for this url: %s"
                    % url
                )

            if html is None:
                raise FatalError("Could not find data for this url: %s" % url)

            scraped = scrape(html)

            if scraped is None:
                raise FatalError(
                    "Could not extract Facebook likes from this url's like button: %s"
                    % url
                )

            enricher.writerow(row, scraped)



import warnings
from contextlib import contextmanager

try:
    from bs4 import XMLParsedAsHTMLWarning
except ImportError:
    XMLParsedAsHTMLWarning = None


@contextmanager
def suppress_xml_parsed_as_html_warnings(bypass=False):
    try:
        if XMLParsedAsHTMLWarning is None or bypass:
            yield
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=XMLParsedAsHTMLWarning)
                yield
    finally:
        pass



# =============================================================================
# Minet Facebook Comments Scraping Functions
# =============================================================================
#
# Functions able to scrape comments from Facebook posts.
#
import re
import sys
import soupsieve
from collections import deque
from urllib.parse import urljoin
from ural import force_protocol, add_query_argument
from ural.facebook import (
    parse_facebook_url,
    convert_facebook_url_to_mobile,
    has_facebook_comments,
    FacebookGroup as ParsedFacebookGroup,
    FacebookHandle as ParsedFacebookHandle,
    FacebookUser as ParsedFacebookUser,
    FacebookPost as ParsedFacebookPost,
    FacebookVideo as ParsedFacebookVideo,
    FacebookPhoto as ParsedFacebookPhoto,
)

from minet.utils import (
    rate_limited_method,
    RateLimiterState,
    clean_human_readable_numbers,
)
from minet.dates import parse_date
from minet.web import (
    create_pool_manager,
    request,
    create_request_retryer,
    retrying_method,
)
from minet.scrape.utils import BeautifulSoupWithoutXHTMLWarnings
from minet.scrape.std import get_display_text
from minet.facebook.utils import grab_facebook_cookie
from minet.facebook.formatters import (
    FacebookComment,
    FacebookPost,
    FacebookUser,
    FacebookPostWithReaction,
)
from minet.facebook.exceptions import (
    FacebookInvalidCookieError,
    FacebookInvalidTargetError,
    FacebookNotPostError,
    FacebookWatchError,
)
from minet.facebook.constants import (
    FACEBOOK_MOBILE_DEFAULT_THROTTLE,
    FACEBOOK_MOBILE_URL,
)

VALID_ID_RE = re.compile(r"^\d+$")


def convert_url_to_mobile(url):
    url = force_protocol(url, "https")
    return convert_facebook_url_to_mobile(url)


def cleanup_post_link(url):
    url = url.replace("//m.", "//www.")

    return url.split("?", 1)[0]


def extract_user_information_from_link(element):
    user_label = element.get_text().strip()
    user_href = element.get("href")
    user = parse_facebook_url(resolve_relative_url(user_href))

    return user_label, user


def resolve_relative_url(url):
    return urljoin(FACEBOOK_MOBILE_URL, url)


def scrape_comments(html, direction=None, in_reply_to=None):
    soup = BeautifulSoupWithoutXHTMLWarnings(html, "lxml")

    data = {
        "direction": direction,
        "post_id": None,
        "comments": [],
        "next": None,
        "replies": [],
        "in_reply_to": in_reply_to,
    }

    # Detecting if we are in a video pagelet
    video_pagelet = soup.select_one("#mobile_injected_video_feed_pagelet")

    if video_pagelet is not None:
        actual_comments_link = video_pagelet.select_one('a[href^="/story.php?"]')

        if actual_comments_link:
            data["next"] = resolve_relative_url(actual_comments_link.get("href"))

        return data

    if not in_reply_to:
        if direction is None or direction == "forward":
            next_link = soup.select_one('[id^="see_next_"] > a[href]')

            if next_link:
                data["next"] = resolve_relative_url(next_link.get("href"))

                if direction is None:
                    data["direction"] = "forward"

        if direction is None or direction == "backward":
            next_link = soup.select_one('[id^="see_prev_"] > a[href]')

            if next_link:
                data["next"] = resolve_relative_url(next_link.get("href"))

                if direction is None:
                    data["direction"] = "backward"
    else:
        if direction is None or direction == "backward":
            next_link = soup.select_one('[id^="comment_replies_more_1"] > a[href]')

            if next_link:
                data["next"] = resolve_relative_url(next_link.get("href"))

                if direction is None:
                    data["direction"] = "backward"

    valid_items = (
        item
        for item in soup.select("[id]:has(h3 > a)")
        if VALID_ID_RE.match(item.get("id"))
        and not item.parent.get("id", "").startswith("comment_replies_more")
    )

    for item in valid_items:
        item_id = item.get("id")

        # Skipping comment if same as commented
        if item_id == in_reply_to:
            continue

        user_label, user = extract_user_information_from_link(item.select_one("h3 > a"))

        # TODO: link to comment
        content_elements_candidates = item.select_one("h3").find_next_siblings("div")
        content_elements = []
        content_elements_html = []

        for el in content_elements_candidates:
            if el.select_one("[id^=like_]"):
                break

            content_elements_html.append(el)

            if el.get_text().strip():
                content_elements.append(el)

        comment_text = get_display_text(content_elements)
        comment_html = "".join(str(el) for el in content_elements_html)

        formatted_date = item.select_one("abbr").get_text().strip()
        parsed_date = parse_date(formatted_date)

        post_id_item = item.select_one('[id^="like_"]')

        if post_id_item is None:
            raise TypeError

        post_id = item.select_one('[id^="like_"]').get("id").split("_")[1]

        # NOTE: this could be better (we already know this beforehand)
        data["post_id"] = post_id

        reactions_item = item.select_one('[href^="/ufi/reaction/"]')
        reactions = "0"

        if reactions_item is not None:
            reactions = reactions_item.get_text().strip()

        replies_items = item.select('a[href^="/comment/replies"]')
        replies = "0"

        if len(replies_items) > 0:
            replies_item = replies_items[-1]

            if replies_item is not None:
                replies_text = replies_item.get_text()

                if replies_text != "Reply":

                    if "See all" in replies_text:
                        replies = (
                            replies_text.split("See all")[-1]
                            .split(" replies")[0]
                            .strip()
                        )
                    else:
                        replies = replies_text.split("·")[-1].split(" repl")[0].strip()

                    replies_url = replies_item.get("href")
                    data["replies"].append((resolve_relative_url(replies_url), item_id))

        data["comments"].append(
            FacebookComment(
                post_id=post_id,
                id=item_id,
                user_id=getattr(user, "id", ""),
                user_handle=getattr(user, "handle", ""),
                user_url=getattr(user, "url", ""),
                user_label=user_label,
                text=comment_text,
                html=comment_html,
                formatted_date=formatted_date,
                date=parsed_date,
                reactions=reactions,
                replies=replies,
                in_reply_to=in_reply_to,
            )
        )

    return data


def scrape_posts(html):
    soup = BeautifulSoupWithoutXHTMLWarnings(html, "lxml")

    next_link = soup.select_one('a[href*="?bacr="], a[href*="&bacr="]')

    if next_link is not None:
        next_link = resolve_relative_url(next_link.get("href"))

    posts = []
    post_elements = soup.select("#m_group_stories_container > div > [data-ft]")

    for el in post_elements:
        full_story_link = soupsieve.select_one('a:-soup-contains("Full Story")', el)

        if not full_story_link:
            continue

        post_url = cleanup_post_link(full_story_link.get("href"))

        user_label, user = extract_user_information_from_link(el.select_one("h3 a"))

        formatted_date = el.select_one("abbr").get_text().strip()
        parsed_date = parse_date(formatted_date)

        reactions_item = el.select_one('[id^="like_"]')
        reactions = "0"

        if reactions_item:
            reactions_text = reactions_item.get_text()

            if reactions_text.count("·") > 1:
                reactions = reactions_text.split("·", 1)[0].strip()

        comments_item = soupsieve.select_one('a:-soup-contains(" Comment")', el)
        comments = "0"

        if comments_item:
            comments = comments_item.get_text().split("Comment", 1)[0].strip()

        text_root = el.select_one('[data-ft=\'{"tn":"*s"}\']')
        additional_html_roots = []

        img_root = el.select_one('[data-ft=\'{"tn":"H"}\']')

        if img_root:
            additional_html_roots.append(img_root)

        all_text_elements = text_root.find_all("div", recursive=False)

        text_elements = []
        translated_text_elements = []
        translation_link = None

        for text_el in all_text_elements:
            translation_link = text_el.select_one(
                'a[href^="/basic/translation_preferences/"]'
            )
            if translation_link is None:
                text_elements.append(text_el)
            else:
                translation_link.extract()
                translated_text_elements.append(text_el)

        html_elements = text_elements + additional_html_roots

        comment_text = get_display_text(text_elements)
        comment_html = "".join(str(el) for el in html_elements)

        translated_comment_text = get_display_text(translated_text_elements)
        translated_comment_html = "".join(str(el) for el in translated_text_elements)
        translated_from = (
            translation_link.get_text().rsplit("from ", 1)[-1].strip()
            if translation_link
            else None
        )

        post = FacebookPost(
            url=post_url,
            user_id=getattr(user, "id", ""),
            user_handle=getattr(user, "handle", ""),
            user_url=getattr(user, "url", ""),
            user_label=user_label,
            text=comment_text,
            html=comment_html,
            translated_text=translated_comment_text,
            translated_html=translated_comment_html,
            translated_from=translated_from,
            formatted_date=formatted_date,
            date=parsed_date,
            reactions=reactions,
            comments=comments,
        )

        posts.append(post)

    return next_link, posts


def scrape_video(soup):

    video_url_html = soup.select_one("meta[property='og:video:url']")

    if not video_url_html:

        title = soup.head.select_one("title").get_text().lower()

        if title == "facebook watch":
            raise FacebookWatchError

        raise FacebookInvalidTargetError

    video_url = video_url_html.get("content")

    user_label = None
    user = None
    user_link = soup.select_one("h3 a")

    if user_link:
        user_label, user = extract_user_information_from_link(user_link)

    formatted_date = soup.select_one("abbr").get_text().strip()
    parsed_date = parse_date(formatted_date)

    reactions_item = soup.select_one('[id^="like_"]')
    reactions = "0"

    if reactions_item:
        reactions_text = reactions_item.get_text()

        if reactions_text.count("·") > 1:
            reactions = clean_human_readable_numbers(
                reactions_text.split("·", 1)[0].strip()
            )

    comments_item = soupsieve.select_one('a:-soup-contains(" Comment")', soup)
    comments = "0"

    if comments_item:
        comments = comments_item.get_text().split("Comment", 1)[0].strip()

    text_root = soup.select_one('[data-ft=\'{"tn":"*s"}\']')
    additional_html_roots = []

    img_root = soup.select_one('[data-ft=\'{"tn":"H"}\']')

    if img_root:
        additional_html_roots.append(img_root)

    all_text_elements = text_root.find_all("div", recursive=False)

    text_elements = []
    translated_text_elements = []
    translation_link = None

    for text_el in all_text_elements:
        translation_link = text_el.select_one(
            'a[href^="/basic/translation_preferences/"]'
        )
        if translation_link is None:
            text_elements.append(text_el)
        else:
            translation_link.extract()
            translated_text_elements.append(text_el)

    html_elements = text_elements + additional_html_roots

    comment_text = get_display_text(text_elements)
    comment_html = "".join(str(el) for el in html_elements)

    translated_comment_text = get_display_text(translated_text_elements)
    translated_comment_html = "".join(str(el) for el in translated_text_elements)
    translated_from = (
        translation_link.get_text().rsplit("from ", 1)[-1].strip()
        if translation_link
        else None
    )

    post = FacebookPostWithReaction(
        url=video_url,
        user_id=getattr(user, "id", ""),
        user_handle=getattr(user, "handle", ""),
        user_url=getattr(user, "url", ""),
        user_label=user_label,
        text=comment_text,
        html=comment_html,
        translated_text=translated_comment_text,
        translated_html=translated_comment_html,
        translated_from=translated_from,
        formatted_date=formatted_date,
        date=parsed_date,
        reactions=reactions,
        comments=comments,
        reactions_types=None,
    )
    return post


def scrape_photo(soup):

    photo_url_html = soup.select_one("meta[property='og:url']")

    if not photo_url_html:

        raise FacebookInvalidTargetError

    photo_url = photo_url_html.get("content")

    user_label = None
    user = None
    user_link = soup.select_one("a[class='actor-link']")

    if user_link:
        user_label, user = extract_user_information_from_link(user_link)

    formatted_date = soup.select_one("abbr").get_text().strip()
    parsed_date = parse_date(formatted_date)

    reactions_item = soup.select_one("a[href^='/ufi/reaction/profile/browser']")

    if reactions_item:
        reactions = clean_human_readable_numbers(reactions_item.get_text())
        reactions_types = []

        for reaction in reactions_item.find("div").select("img"):
            reactions_types.append(reaction["alt"])
        reactions_types = "|".join(reactions_types)

    comments = None

    text_root = soup.select_one(".msg")

    additional_html_roots = []

    img_root = soup.select_one("img[class='img']").get("src")

    if img_root:
        additional_html_roots.append(img_root)

    all_text_elements = (
        text_root.find_all("div", recursive=False) if text_root else None
    )

    text_elements = []
    translated_text_elements = []
    translation_link = None

    if all_text_elements:
        for text_el in all_text_elements:
            translation_link = text_el.select_one(
                'a[href^="/basic/translation_preferences/"]'
            )
            if translation_link is None:
                text_elements.append(text_el)
            else:
                translation_link.extract()
                translated_text_elements.append(text_el)

    html_elements = text_elements + additional_html_roots

    comment_text = get_display_text(text_elements) if text_elements else None
    comment_html = "".join(str(el) for el in html_elements)

    translated_comment_text = get_display_text(translated_text_elements)
    translated_comment_html = "".join(str(el) for el in translated_text_elements)
    translated_from = (
        translation_link.get_text().rsplit("from ", 1)[-1].strip()
        if translation_link
        else None
    )

    post = FacebookPostWithReaction(
        url=photo_url,
        user_id=getattr(user, "id", ""),
        user_handle=getattr(user, "handle", ""),
        user_url=getattr(user, "url", ""),
        user_label=user_label,
        text=comment_text,
        html=comment_html,
        translated_text=translated_comment_text,
        translated_html=translated_comment_html,
        translated_from=translated_from,
        formatted_date=formatted_date,
        date=parsed_date,
        reactions=reactions,
        comments=comments,
        reactions_types=reactions_types,
    )
    return post


def scrape_post(html):
    soup = BeautifulSoupWithoutXHTMLWarnings(html, "lxml")

    # with open("./dump.html", "w") as f:
    #     f.write(html)

    if soup.find("div", id="MPhotoContent"):
        return scrape_photo(soup)

    post_url_htlm = soup.head.select_one("meta[property='og:url']")

    if not post_url_htlm:

        title = soup.head.select_one("title").get_text().lower()

        if title == "facebook watch":
            raise FacebookWatchError

        raise FacebookInvalidTargetError

    post_url = post_url_htlm.get("content") if post_url_htlm else None

    if post_url == "":
        return scrape_video(soup)

    user_label = None
    user = None
    user_link = soup.select_one("h3 a")
    if user_link:
        user_label, user = extract_user_information_from_link(user_link)

    formatted_date = soup.select_one("abbr").get_text().strip()
    parsed_date = parse_date(formatted_date)

    reactions_item = soup.select_one("a[href^='/ufi/reaction/profile/browser']")

    if reactions_item:
        reactions = clean_human_readable_numbers(reactions_item.get_text())
        reactions_types = []

        for reaction in reactions_item.find("div").select("img"):
            reactions_types.append(reaction["alt"])
        reactions_types = "|".join(reactions_types)

    comments = None

    text_root = soup.select_one('[data-ft=\'{"tn":"*s"}\']')
    additional_html_roots = []

    img_root = soup.select_one('[data-ft=\'{"tn":"H"}\']')

    if img_root:
        additional_html_roots.append(img_root)

    all_text_elements = (
        text_root.find_all("div", recursive=False) if text_root else None
    )

    text_elements = []
    translated_text_elements = []
    translation_link = None

    if all_text_elements:
        for text_el in all_text_elements:
            translation_link = text_el.select_one(
                'a[href^="/basic/translation_preferences/"]'
            )
            if translation_link is None:
                text_elements.append(text_el)
            else:
                translation_link.extract()
                translated_text_elements.append(text_el)

    html_elements = text_elements + additional_html_roots

    comment_text = get_display_text(text_elements) if text_elements else None
    comment_html = "".join(str(el) for el in html_elements)

    translated_comment_text = get_display_text(translated_text_elements)
    translated_comment_html = "".join(str(el) for el in translated_text_elements)
    translated_from = (
        translation_link.get_text().rsplit("from ", 1)[-1].strip()
        if translation_link
        else None
    )

    post = FacebookPostWithReaction(
        url=post_url,
        user_id=getattr(user, "id", ""),
        user_handle=getattr(user, "handle", ""),
        user_url=getattr(user, "url", ""),
        user_label=user_label,
        text=comment_text,
        html=comment_html,
        translated_text=translated_comment_text,
        translated_html=translated_comment_html,
        translated_from=translated_from,
        formatted_date=formatted_date,
        date=parsed_date,
        reactions=reactions,
        comments=comments,
        reactions_types=reactions_types,
    )
    return post


class FacebookMobileScraper(object):
    def __init__(self, cookie, throttle=FACEBOOK_MOBILE_DEFAULT_THROTTLE):

        # Grabbing cookie
        self.cookie = grab_facebook_cookie(cookie)

        if self.cookie is None:
            raise FacebookInvalidCookieError(target=cookie)

        self.pool_manager = create_pool_manager()

        self.rate_limiter_state = RateLimiterState(1, throttle)
        self.retryer = create_request_retryer()

    @rate_limited_method()
    @retrying_method()
    def request_page(self, url):
        url = add_query_argument(url, "locale", "en_US")

        response = request(
            url,
            pool_manager=self.pool_manager,
            cookie=self.cookie,
            headers={"User-Agent": "curl/7.68.0", "Accept-Language": "en"},
            known_encoding="utf-8",
        )

        return response.text()

    def comments(self, url, detailed=False, per_call=False):

        if not has_facebook_comments(url):
            raise FacebookInvalidTargetError

        # Reformatting url to hit mobile website
        url = convert_url_to_mobile(url)

        def generator():
            url_queue = deque([(url, None, None)])

            calls = 0
            replies = 0

            while len(url_queue) != 0:
                current_url, direction, in_reply_to = url_queue.popleft()

                html = self.request_page(current_url)
                try:
                    data = scrape_comments(html, direction, in_reply_to)
                except TypeError:
                    # with open('./dump.html', 'w') as f:
                    #     f.write(html)
                    print(
                        "Could not process comment in %s" % current_url, file=sys.stderr
                    )
                    return

                calls += 1

                for reply_url, commented_id in data["replies"]:
                    url_queue.append((reply_url, None, commented_id))

                if data["next"] is not None:
                    url_queue.append((data["next"], data["direction"], in_reply_to))

                comments = data["comments"]

                for comment in data["comments"]:
                    if in_reply_to is not None:
                        replies += 1

                    if not per_call:
                        yield comment

                if per_call and len(comments) > 0:
                    if detailed:
                        details = {
                            "calls": calls,
                            "replies": replies,
                            "queue_size": len(url_queue),
                        }

                        yield details, comments
                    else:
                        yield comments

        return generator()

    def posts(self, url):
        parsed = parse_facebook_url(url)

        if not isinstance(parsed, ParsedFacebookGroup):
            raise FacebookInvalidTargetError

        url = convert_url_to_mobile(parsed.url)

        def generator():
            current_url = url

            while True:
                html = self.request_page(current_url)
                # with open("./dump.html", "w") as f:
                #     f.write(html)

                next_url, posts = scrape_posts(html)

                for post in posts:
                    yield post

                if next_url is None or len(posts) == 0:
                    break

                current_url = next_url

        return generator()

    def post(self, url):
        parsed = parse_facebook_url(url)

        if (
            not isinstance(parsed, ParsedFacebookPost)
            and not isinstance(parsed, ParsedFacebookVideo)
            and not isinstance(parsed, ParsedFacebookPhoto)
        ):
            raise FacebookNotPostError

        url = convert_url_to_mobile(parsed.url)

        html = self.request_page(url)

        return scrape_post(html)

    def post_author(self, url):
        if not has_facebook_comments(url):
            raise FacebookInvalidTargetError

        # Reformatting url to hit mobile website
        url = convert_url_to_mobile(url)

        html = self.request_page(url)
        soup = BeautifulSoupWithoutXHTMLWarnings(html, "lxml")

        user_item = soup.select_one("[data-ft] h3 a[href]")

        if user_item is None:
            return None

        parsed = parse_facebook_url(user_item.get("href"), allow_relative_urls=True)
        user_label = user_item.get_text().strip()

        if isinstance(parsed, ParsedFacebookHandle):
            return FacebookUser(user_label, None, parsed.handle, parsed.url)
        elif isinstance(parsed, ParsedFacebookUser):
            return FacebookUser(user_label, parsed.id, parsed.handle, parsed.url)
        else:
            raise TypeError



# =============================================================================
# Minet Telegram Formatters
# =============================================================================
#
# Various formatters for Telegram data.
#
from casanova import namedrecord

from minet.telegram.constants import (
    TELEGRAM_INFOS_CSV_HEADERS,
    TELEGRAM_MESSAGES_CSV_HEADERS,
)

TelegramChannelInfos = namedrecord("TelegramChannelInfos", TELEGRAM_INFOS_CSV_HEADERS)
TelegramChannelMessages = namedrecord(
    "TelegramChannelMessages", TELEGRAM_MESSAGES_CSV_HEADERS
)



# =============================================================================
# Minet Youtube Utils
# =============================================================================
#
# Miscellaneous helper functions related to the Youtube API.
#
from pytz import timezone
from datetime import datetime
from ural import is_url
from ural.youtube import (
    is_youtube_video_id,
    extract_video_id_from_youtube_url,
    parse_youtube_url,
    YoutubeChannel,
)


def seconds_to_midnight_pacific_time():
    now_utc = timezone("utc").localize(datetime.utcnow())
    pacific_time = now_utc.astimezone(timezone("US/Pacific")).replace(tzinfo=None)
    midnight_pacific = datetime.combine(pacific_time, datetime.min.time())

    return (midnight_pacific - pacific_time).seconds


def ensure_video_id(target):
    if is_youtube_video_id(target):
        return target

    return extract_video_id_from_youtube_url(target)


def ensure_channel_id(target):
    if not is_url(target):
        return False, target

    parsed = parse_youtube_url(target)

    if parsed is None or not isinstance(parsed, YoutubeChannel):
        return False, None

    if parsed.id:
        return False, parsed.id

    return True, None


def get_channel_main_playlist_id(channel_id):
    return "UU" + channel_id[2:]



# =============================================================================
# Minet Twitter User Tweets CLI Action
# =============================================================================
#
# Logic of the `tw user-tweets` action.
#
from datetime import datetime
from twitwi import normalize_tweet, normalize_tweets_payload_v2, format_tweet_as_csv_row
from twitwi.constants import TWEET_FIELDS, TWEET_EXPANSIONS, TWEET_PARAMS
from twitter import TwitterHTTPError

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.exceptions import FatalError
from minet.twitter.constants import TWITTER_API_MAX_STATUSES_COUNT
from minet.cli.twitter.utils import (
    is_not_user_id,
    is_probably_not_user_screen_name,
    with_twitter_client,
)

ITEMS_PER_PAGE = 100


@with_enricher_and_loading_bar(
    headers=TWEET_FIELDS,
    title="Retrieving user tweets",
    unit="users",
    nested=True,
    sub_unit="tweets",
)
@with_twitter_client()
def action(cli_args, client, enricher, loading_bar):
    for row, user in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(user):
            max_id = None

            min_date_reached = False

            kwargs = {}

            while True:
                if cli_args.v2:
                    if is_not_user_id(user):
                        raise FatalError(
                            "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. With --api-v2, you can only use user ids to retrieve followers."
                        )

                    kwargs["max_results"] = ITEMS_PER_PAGE
                    kwargs["params"] = TWEET_PARAMS
                    kwargs["expansions"] = ",".join(TWEET_EXPANSIONS)

                    if cli_args.exclude_retweets:
                        kwargs["exclude"] = "retweets"
                    if cli_args.min_date:
                        utc_dt = datetime.fromtimestamp(cli_args.min_date).strftime(
                            "%Y-%m-%d"
                        )
                        kwargs["start_time"] = utc_dt + "T00:00:00Z"

                    try:
                        tweets = client.call(["users", user, "tweets"], **kwargs)
                    except TwitterHTTPError as e:
                        if e.e.code == 404:
                            loading_bar.print('Could not find user "%s"' % user)
                        else:
                            loading_bar.print(
                                'An error happened when attempting to retrieve tweets from "%s" (HTTP status %i)'
                                % (user, e.e.code)
                            )

                        break

                    try:
                        normalized_tweets = normalize_tweets_payload_v2(
                            tweets, locale=cli_args.timezone, collection_source="api"
                        )
                    except TypeError:
                        loading_bar.print(
                            'An error happened when attempting to retrieve tweets from "%s"'
                            % user
                        )
                        loading_bar.print(
                            'details: "%s"' % tweets["errors"][0]["detail"]
                        )
                        break

                    for normalized_tweet in normalized_tweets:
                        addendum = format_tweet_as_csv_row(normalized_tweet)
                        enricher.writerow(row, addendum)

                    if "next_token" in tweets["meta"]:
                        kwargs["pagination_token"] = tweets["meta"]["next_token"]
                    else:
                        break

                else:
                    if cli_args.ids:
                        if is_not_user_id(user):
                            raise FatalError(
                                "The column given as argument doesn't contain user ids, you have probably given user screen names as argument instead. \nTry removing --ids from the command."
                            )

                        kwargs = {"user_id": user}

                    else:
                        if is_probably_not_user_screen_name(user):
                            raise FatalError(
                                "The column given as argument probably doesn't contain user screen names, you have probably given user ids as argument instead. \nTry adding --ids to the command."
                            )
                            # force flag to add

                        kwargs = {"screen_name": user}

                    kwargs["include_rts"] = not cli_args.exclude_retweets
                    kwargs["count"] = TWITTER_API_MAX_STATUSES_COUNT
                    kwargs["tweet_mode"] = "extended"

                    if max_id is not None:
                        kwargs["max_id"] = max_id

                    try:
                        tweets = client.call(["statuses", "user_timeline"], **kwargs)
                    except TwitterHTTPError as e:
                        if e.e.code == 404:
                            loading_bar.print('Could not find user "%s"' % user)
                        else:
                            loading_bar.print(
                                'An error happened when attempting to retrieve tweets from "%s" (HTTP status %i)'
                                % (user, e.e.code)
                            )

                        break

                    if not tweets:
                        break

                    max_id = min(int(tweet["id_str"]) for tweet in tweets) - 1

                    for tweet in tweets:
                        tweet = normalize_tweet(
                            tweet, locale=cli_args.timezone, collection_source="api"
                        )
                        addendum = format_tweet_as_csv_row(tweet)

                        if cli_args.min_date:
                            if int(addendum[1]) < cli_args.min_date:
                                min_date_reached = True
                                break

                        enricher.writerow(row, addendum)
                        loading_bar.nested_advance()

                if min_date_reached:
                    break



# =============================================================================
# Minet Spiders
# =============================================================================
#
#
from typing import (
    cast,
    Any,
    Optional,
    List,
    Tuple,
    TypeVar,
    Union,
    Callable,
    Dict,
    Iterator,
    Iterable,
    Generic,
)
from minet.types import TypedDict, NotRequired

from urllib.parse import urljoin
from bs4 import BeautifulSoup

from minet.crawl.types import (
    UrlOrCrawlJob,
    CrawlJob,
    CrawlJobDataType,
    CrawlResultDataType,
)
from minet.types import AnyFileTarget
from minet.fs import load_definition
from minet.scrape import Scraper
from minet.web import Response
from minet.utils import PseudoFStringFormatter

FORMATTER = PseudoFStringFormatter()

T = TypeVar("T")


def ensure_list(value: Union[T, List[T]]) -> List[T]:
    if not isinstance(value, list):
        return [value]
    return value


SpiderNextJobs = Optional[Iterable[Union[str, CrawlJob[CrawlJobDataType]]]]
SpiderResult = Tuple[CrawlResultDataType, SpiderNextJobs[CrawlJobDataType]]


class Spider(Generic[CrawlJobDataType, CrawlResultDataType]):
    START_URL: Optional[str] = None
    START_URLS: Optional[List[str]] = None

    def start_jobs(self) -> Optional[Iterable[UrlOrCrawlJob[CrawlJobDataType]]]:
        return None

    def __call__(
        self, job: CrawlJob[CrawlJobDataType], response: Response
    ) -> SpiderResult[CrawlResultDataType, CrawlJobDataType]:
        raise NotImplementedError

    def __repr__(self):
        class_name = self.__class__.__name__

        return "<%(class_name)s>" % {"class_name": class_name}


FunctionSpiderCallable = Callable[
    [CrawlJob[CrawlJobDataType], Response],
    SpiderResult[CrawlResultDataType, CrawlJobDataType],
]


class FunctionSpider(Spider[CrawlJobDataType, CrawlResultDataType]):
    fn: FunctionSpiderCallable[CrawlJobDataType, CrawlResultDataType]

    def __init__(
        self,
        fn: FunctionSpiderCallable[CrawlJobDataType, CrawlResultDataType],
        start_jobs: Optional[Iterable[UrlOrCrawlJob[CrawlJobDataType]]] = None,
    ):
        self.fn = fn
        self.__start_jobs = start_jobs

    def start_jobs(self) -> Optional[Iterable[UrlOrCrawlJob[CrawlJobDataType]]]:
        return self.__start_jobs

    def __call__(
        self, job: CrawlJob[CrawlJobDataType], response: Response
    ) -> SpiderResult[CrawlResultDataType, CrawlJobDataType]:
        return self.fn(job, response)


class DefinitionSpiderOutput(Generic[CrawlResultDataType]):
    __slots__ = ("default", "named")

    def __init__(self):
        self.default = None
        self.named = {}


class DefinitionSpiderTarget(TypedDict, Generic[CrawlJobDataType]):
    url: str
    spider: NotRequired[str]
    data: NotRequired[CrawlJobDataType]


class DefinitionSpider(
    Spider[CrawlJobDataType, DefinitionSpiderOutput[CrawlResultDataType]]
):
    definition: Dict[str, Any]
    next_definition: Optional[Dict[str, Any]]
    max_depth: int
    scraper: Optional[Scraper]
    scrapers: Dict[str, Scraper]
    next_scraper: Optional[Scraper]
    next_scrapers: Dict[str, Scraper]

    def __init__(self, definition: Union[AnyFileTarget, Dict[str, Any]]):

        if not isinstance(definition, dict):
            definition = load_definition(definition)

        definition = cast(Dict[str, Any], definition)

        # Descriptors
        self.definition = definition
        self.next_definition = definition.get("next")

        # Settings
        self.max_depth = definition.get("max_depth", float("inf"))

        # Scrapers
        self.scraper = None
        self.scrapers = {}
        self.next_scraper = None
        self.next_scrapers = {}

        if "scraper" in definition:
            self.scraper = Scraper(definition["scraper"])

        if "scrapers" in definition:
            for name, scraper in definition["scrapers"].items():
                self.scrapers[name] = Scraper(scraper)

        if self.next_definition is not None:
            if "scraper" in self.next_definition:
                self.next_scraper = Scraper(self.next_definition["scraper"])

            if "scrapers" in self.next_definition:
                for name, scraper in self.next_definition["scrapers"].items():
                    self.next_scrapers[name] = Scraper(scraper)

    def start_jobs(self) -> Iterable[UrlOrCrawlJob[CrawlJobDataType]]:

        # TODO: possibility to name this as jobs and get additional metadata
        start_urls = ensure_list(self.definition.get("start_url", [])) + ensure_list(
            self.definition.get("start_urls", [])
        )

        for url in start_urls:
            yield url

    def __scrape(
        self, job: CrawlJob[CrawlJobDataType], response: Response, soup: BeautifulSoup
    ) -> DefinitionSpiderOutput[CrawlResultDataType]:
        scraped = DefinitionSpiderOutput[CrawlResultDataType]()

        context = {"url": job.url}

        if self.scraper is not None:
            scraped.default = self.scraper(soup, context=context)

        for name, scraper in self.scrapers.items():
            scraped.named[name] = scraper(soup, context=context)

        return scraped

    def __job_from_target(
        self,
        current_url: str,
        target: Union[str, DefinitionSpiderTarget[CrawlJobDataType]],
    ) -> CrawlJob[CrawlJobDataType]:
        if isinstance(target, str):
            return CrawlJob(url=urljoin(current_url, target))

        else:

            # TODO: validate target
            return CrawlJob(
                url=urljoin(current_url, target["url"]),
                spider=target.get("spider"),
                data=target.get("data"),
            )

    def __next_targets(
        self, response: Response, soup: BeautifulSoup
    ) -> Iterator[Union[str, DefinitionSpiderTarget[CrawlJobDataType]]]:

        # Scraping next results
        if self.next_scraper is not None:
            scraped = self.next_scraper(soup)

            if scraped is not None:
                if isinstance(scraped, list):
                    yield from scraped
                else:
                    yield scraped

        if self.next_scrapers:
            for scraper in self.next_scrapers.values():
                scraped = scraper(soup)

                if scraped is not None:
                    if isinstance(scraped, list):
                        yield from scraped
                    else:
                        yield scraped

        # Formatting next url
        if self.next_definition is not None and "format" in self.next_definition:
            yield FORMATTER.format(self.next_definition["format"])

    def __next_jobs(
        self, job: CrawlJob[CrawlJobDataType], response: Response, soup: BeautifulSoup
    ):
        if not self.next_definition:
            return

        assert job.depth is not None

        next_depth = job.depth + 1

        if next_depth > self.max_depth:
            return

        for target in self.__next_targets(response, soup):
            yield self.__job_from_target(response.url, target)

    def __call__(
        self, job: CrawlJob[CrawlJobDataType], response: Response
    ) -> SpiderResult[DefinitionSpiderOutput[CrawlResultDataType], CrawlJobDataType]:
        soup = response.soup()

        return self.__scrape(job, response, soup), self.__next_jobs(job, response, soup)



# =============================================================================
# Minet Youtube Channel Videos CLI Action
# =============================================================================
#
# Action reading an input CSV file line by line and retrieving videos from
# the given Youtube channels using Google's APIs.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.youtube import YouTubeAPIClient
from minet.youtube.constants import YOUTUBE_PLAYLIST_VIDEO_SNIPPET_CSV_HEADERS


@with_enricher_and_loading_bar(
    headers=YOUTUBE_PLAYLIST_VIDEO_SNIPPET_CSV_HEADERS,
    title="Retrieving channel videos",
    unit="channels",
    sub_unit="videos",
    nested=True,
)
def action(cli_args, enricher, loading_bar):
    client = YouTubeAPIClient(cli_args.key)

    for row, channel_id in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(channel_id):

            for video in client.channel_videos(channel_id):
                enricher.writerow(row, video.as_csv_row())
                loading_bar.nested_advance()
